import os
from chester import config
from chester.utils_logger import timelog
import psutil
import os
import glob
import pickle
import time


def check_available_nodes():
    all_stats_path = glob.glob(config.GPU_STATE_DIR + '/*.pkl')
    all_stats_path.sort()
    stats_names = [p.split('/')[-1][:-4] for p in all_stats_path]
    all_stats = []
    for p in all_stats_path:
        while 1:
            try:
                all_stats.append(pickle.load(open(p, 'rb')))
                time.sleep(0.1)
            except:
                print(p, ' is broken')
                continue
            break
    available_nodes = {}
    print("    Num process |   Node name")
    for name, (sys_data, gpu_data) in zip(stats_names, all_stats):
        counter = 0
        for i, data in enumerate(gpu_data):
            for proc_data in data['procs']:
                if proc_data[1] == 'zixuanhu':
                    counter += 1
        print(f"{counter:15} | {name}")
    return available_nodes


if __name__ == '__main__':
    check_available_nodes()



import boto3
import re
import sys
import json
import botocore
import os
from rllab.misc import console
from string import Template
import os.path as osp

CHESTER_DIR = osp.dirname(__file__)
ACCESS_KEY = os.environ["AWS_ACCESS_KEY"]
ACCESS_SECRET = os.environ["AWS_ACCESS_SECRET"]
S3_BUCKET_NAME = os.environ["RLLAB_S3_BUCKET"]

ALL_REGION_AWS_SECURITY_GROUP_IDS = {}
ALL_REGION_AWS_KEY_NAMES = {}

CONFIG_TEMPLATE = Template("""
import os.path as osp
import os

PROJECT_PATH = osp.abspath(osp.join(osp.dirname(__file__), '..'))

AWS_NETWORK_INTERFACES = []

MUJOCO_KEY_PATH = osp.expanduser("~/.mujoco")

USE_GPU = False

USE_TF = True

AWS_REGION_NAME = "us-east-2"

if USE_GPU:
    DOCKER_IMAGE = "dementrock/rllab3-shared-gpu"
else:
    DOCKER_IMAGE = "dementrock/rllab3-shared"

DOCKER_LOG_DIR = "/tmp/expt"

AWS_S3_PATH = "s3://$s3_bucket_name/rllab/experiments"

AWS_CODE_SYNC_S3_PATH = "s3://$s3_bucket_name/rllab/code"

ALL_REGION_AWS_IMAGE_IDS = {
    "ap-northeast-1": "ami-002f0167",
    "ap-northeast-2": "ami-590bd937",
    "ap-south-1": "ami-77314318",
    "ap-southeast-1": "ami-1610a975",
    "ap-southeast-2": "ami-9dd4ddfe",
    "eu-central-1": "ami-63af720c",
    "eu-west-1": "ami-41484f27",
    "sa-east-1": "ami-b7234edb",
    "us-east-1": "ami-83f26195",
    "us-east-2": "ami-66614603",
    "us-west-1": "ami-576f4b37",
    "us-west-2": "ami-b8b62bd8"
}

AWS_IMAGE_ID = ALL_REGION_AWS_IMAGE_IDS[AWS_REGION_NAME]

if USE_GPU:
    AWS_INSTANCE_TYPE = "g2.2xlarge"
else:
    AWS_INSTANCE_TYPE = "c4.4xlarge"

ALL_REGION_AWS_KEY_NAMES = $all_region_aws_key_names

AWS_KEY_NAME = ALL_REGION_AWS_KEY_NAMES[AWS_REGION_NAME]

AWS_SPOT = True

AWS_SPOT_PRICE = '0.5'

AWS_ACCESS_KEY = os.environ.get("AWS_ACCESS_KEY", None)

AWS_ACCESS_SECRET = os.environ.get("AWS_ACCESS_SECRET", None)

AWS_IAM_INSTANCE_PROFILE_NAME = "rllab"

AWS_SECURITY_GROUPS = ["rllab-sg"]

ALL_REGION_AWS_SECURITY_GROUP_IDS = $all_region_aws_security_group_ids

AWS_SECURITY_GROUP_IDS = ALL_REGION_AWS_SECURITY_GROUP_IDS[AWS_REGION_NAME]

FAST_CODE_SYNC_IGNORES = [
    ".git",
    "data",
    "data/local",
    "data/archive",
    "data/debug",
    "data/s3",
    "data/video",
    "src",
    ".idea",
    ".pods",
    "tests",
    "examples",
    "docs",
    ".idea",
    ".DS_Store",
    ".ipynb_checkpoints",
    "blackbox",
    "blackbox.zip",
    "*.pyc",
    "*.ipynb",
    "scratch-notebooks",
    "conopt_root",
    "private/key_pairs",
]

FAST_CODE_SYNC = True

""")


def setup_iam():
    iam_client = boto3.client(
        "iam",
        aws_access_key_id=ACCESS_KEY,
        aws_secret_access_key=ACCESS_SECRET,
    )
    iam = boto3.resource('iam', aws_access_key_id=ACCESS_KEY, aws_secret_access_key=ACCESS_SECRET)

    # delete existing role if it exists
    try:
        existing_role = iam.Role('rllab')
        existing_role.load()
        # if role exists, delete and recreate
        if not query_yes_no(
                "There is an existing role named rllab. Proceed to delete everything rllab-related and recreate?",
                default="no"):
            sys.exit()
        print("Listing instance profiles...")
        inst_profiles = existing_role.instance_profiles.all()
        for prof in inst_profiles:
            for role in prof.roles:
                print("Removing role %s from instance profile %s" % (role.name, prof.name))
                prof.remove_role(RoleName=role.name)
            print("Deleting instance profile %s" % prof.name)
            prof.delete()
        for policy in existing_role.policies.all():
            print("Deleting inline policy %s" % policy.name)
            policy.delete()
        for policy in existing_role.attached_policies.all():
            print("Detaching policy %s" % policy.arn)
            existing_role.detach_policy(PolicyArn=policy.arn)
        print("Deleting role")
        existing_role.delete()
    except botocore.exceptions.ClientError as e:
        if e.response['Error']['Code'] == 'NoSuchEntity':
            pass
        else:
            raise e

    print("Creating role rllab")
    iam_client.create_role(
        Path='/',
        RoleName='rllab',
        AssumeRolePolicyDocument=json.dumps({'Version': '2012-10-17', 'Statement': [
            {'Action': 'sts:AssumeRole', 'Effect': 'Allow', 'Principal': {'Service': 'ec2.amazonaws.com'}}]})
    )

    role = iam.Role('rllab')
    print("Attaching policies")
    role.attach_policy(PolicyArn='arn:aws:iam::aws:policy/AmazonS3FullAccess')
    role.attach_policy(PolicyArn='arn:aws:iam::aws:policy/ResourceGroupsandTagEditorFullAccess')

    print("Creating inline policies")
    iam_client.put_role_policy(
        RoleName=role.name,
        PolicyName='CreateTags',
        PolicyDocument=json.dumps({
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Effect": "Allow",
                    "Action": ["ec2:CreateTags"],
                    "Resource": ["*"]
                }
            ]
        })
    )
    iam_client.put_role_policy(
        RoleName=role.name,
        PolicyName='TerminateInstances',
        PolicyDocument=json.dumps({
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Sid": "Stmt1458019101000",
                    "Effect": "Allow",
                    "Action": [
                        "ec2:TerminateInstances"
                    ],
                    "Resource": [
                        "*"
                    ]
                }
            ]
        })
    )

    print("Creating instance profile rllab")
    iam_client.create_instance_profile(
        InstanceProfileName='rllab',
        Path='/'
    )
    print("Adding role rllab to instance profile rllab")
    iam_client.add_role_to_instance_profile(
        InstanceProfileName='rllab',
        RoleName='rllab'
    )


def setup_s3():
    print("Creating S3 bucket at s3://%s" % S3_BUCKET_NAME)
    s3_client = boto3.client(
        "s3",
        aws_access_key_id=ACCESS_KEY,
        aws_secret_access_key=ACCESS_SECRET,
    )
    try:
        s3_client.create_bucket(
            ACL='private',
            Bucket=S3_BUCKET_NAME,
            CreateBucketConfiguration={
                'LocationConstraint': 'us-east-2'
            }
        )
    except botocore.exceptions.ClientError as e:
        if e.response['Error']['Code'] == 'BucketAlreadyExists':
            raise ValueError("Bucket %s already exists. Please reconfigure S3_BUCKET_NAME" % S3_BUCKET_NAME) from e
        elif e.response['Error']['Code'] == 'BucketAlreadyOwnedByYou':
            print("Bucket already created by you")
        else:
            raise e
    print("S3 bucket created")


def setup_ec2():
    for region in ["us-east-1", "us-east-2", "us-west-1", "us-west-2"]:
        print("Setting up region %s" % region)

        ec2 = boto3.resource(
            "ec2",
            region_name=region,
            aws_access_key_id=ACCESS_KEY,
            aws_secret_access_key=ACCESS_SECRET,
        )
        ec2_client = boto3.client(
            "ec2",
            region_name=region,
            aws_access_key_id=ACCESS_KEY,
            aws_secret_access_key=ACCESS_SECRET,
        )
        existing_vpcs = list(ec2.vpcs.all())
        assert len(existing_vpcs) >= 1
        vpc = existing_vpcs[0]
        print("Creating security group in VPC %s" % str(vpc.id))
        try:
            security_group = vpc.create_security_group(
                GroupName='rllab-sg', Description='Security group for rllab'
            )
        except botocore.exceptions.ClientError as e:
            if e.response['Error']['Code'] == 'InvalidGroup.Duplicate':
                sgs = list(vpc.security_groups.filter(GroupNames=['rllab-sg']))
                security_group = sgs[0]
            else:
                raise e

        ALL_REGION_AWS_SECURITY_GROUP_IDS[region] = [security_group.id]

        ec2_client.create_tags(Resources=[security_group.id], Tags=[{'Key': 'Name', 'Value': 'rllab-sg'}])
        try:
            security_group.authorize_ingress(FromPort=22, ToPort=22, IpProtocol='tcp', CidrIp='0.0.0.0/0')
        except botocore.exceptions.ClientError as e:
            if e.response['Error']['Code'] == 'InvalidPermission.Duplicate':
                pass
            else:
                raise e
        print("Security group created with id %s" % str(security_group.id))

        key_name = 'rllab-%s' % region
        try:
            print("Trying to create key pair with name %s" % key_name)
            key_pair = ec2_client.create_key_pair(KeyName=key_name)
        except botocore.exceptions.ClientError as e:
            if e.response['Error']['Code'] == 'InvalidKeyPair.Duplicate':
                if not query_yes_no("Key pair with name %s exists. Proceed to delete and recreate?" % key_name, "no"):
                    sys.exit()
                print("Deleting existing key pair with name %s" % key_name)
                ec2_client.delete_key_pair(KeyName=key_name)
                print("Recreating key pair with name %s" % key_name)
                key_pair = ec2_client.create_key_pair(KeyName=key_name)
            else:
                raise e

        key_pair_folder_path = os.path.join(CHESTER_DIR, "private", "key_pairs")
        file_name = os.path.join(key_pair_folder_path, "%s.pem" % key_name)

        print("Saving keypair file")
        console.mkdir_p(key_pair_folder_path)
        with os.fdopen(os.open(file_name, os.O_WRONLY | os.O_CREAT, 0o600), 'w') as handle:
            handle.write(key_pair['KeyMaterial'] + '\n')

        # adding pem file to ssh
        os.system("ssh-add %s" % file_name)

        ALL_REGION_AWS_KEY_NAMES[region] = key_name


def write_config():
    print("Writing config file...")
    content = CONFIG_TEMPLATE.substitute(
        all_region_aws_key_names=json.dumps(ALL_REGION_AWS_KEY_NAMES, indent=4),
        all_region_aws_security_group_ids=json.dumps(ALL_REGION_AWS_SECURITY_GROUP_IDS, indent=4),
        s3_bucket_name=S3_BUCKET_NAME,
    )
    config_personal_file = os.path.join(CHESTER_DIR, "config_ec2.py")
    if os.path.exists(config_personal_file):
        if not query_yes_no("config_ec2.py exists. Override?", "no"):
            sys.exit()
    with open(config_personal_file, "wb") as f:
        f.write(content.encode("utf-8"))


def setup():
    setup_s3()
    setup_iam()
    setup_ec2()
    write_config()


def query_yes_no(question, default="yes"):
    """Ask a yes/no question via raw_input() and return their answer.

    "question" is a string that is presented to the user.
    "default" is the presumed answer if the user just hits <Enter>.
        It must be "yes" (the default), "no" or None (meaning
        an answer is required of the user).

    The "answer" return value is True for "yes" or False for "no".
    """
    valid = {"yes": True, "y": True, "ye": True,
             "no": False, "n": False}
    if default is None:
        prompt = " [y/n] "
    elif default == "yes":
        prompt = " [Y/n] "
    elif default == "no":
        prompt = " [y/N] "
    else:
        raise ValueError("invalid default answer: '%s'" % default)

    while True:
        sys.stdout.write(question + prompt)
        choice = input().lower()
        if default is not None and choice == '':
            return valid[default]
        elif choice in valid:
            return valid[choice]
        else:
            sys.stdout.write("Please respond with 'yes' or 'no' "
                             "(or 'y' or 'n').\n")


if __name__ == "__main__":
    setup()


import glob
import sys
import os
import argparse

if __name__ == '__main__':
    batch_dir = sys.argv[1]
    is_dry = int(sys.argv[2])
    launch_scripts = glob.glob(batch_dir + "/*/slurm_launch")
    print(f"Found {len(launch_scripts)} launch scripts in {batch_dir}")
    for x in launch_scripts:
        print("Launching   {script}".format(script=x))
        if not is_dry:
            os.system("sbatch {script}".format(script=x))



import sys
import os
import argparse

# sys.path.append('.')
from chester import config

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument('host', type=str)
    parser.add_argument('folder', type=str)
    parser.add_argument('--dry', action='store_true', default=False)
    parser.add_argument('--bare', action='store_true', default=False)
    args = parser.parse_args()

    args.folder = args.folder.rstrip('/')
    if args.folder.rfind('/') !=-1:
        local_dir = os.path.join('./data', args.host, args.folder[:args.folder.rfind('/')])
    else:
        local_dir = os.path.join('./data', args.host, args.folder)
    remote_data_dir = os.path.join(config.REMOTE_DIR[args.host], 'data', 'local', args.folder)
    command = """rsync -avzh --delete --progress {host}:{remote_data_dir} {local_dir}""".format(host=args.host,
                                                                                                remote_data_dir=remote_data_dir,
                                                                                                local_dir=local_dir)
    if args.bare:
        command += """ --exclude '*.pkl' --exclude '*.png' --exclude '*.gif' --exclude '*.pth' --exclude '*.pt' --include '*.csv' --include '*.json' --delete"""
    if args.dry:
        print(command)
    else:
        os.system(command)



import os
import subprocess
import argparse

def aws_sync(bucket_name, s3_log_dir, target_dir, args):
    cmd = 'aws s3 cp --recursive s3://%s/%s %s' % (bucket_name, s3_log_dir, target_dir)
    exlus = ['"*.pkl"', '"*.gif"', '"*.png"', '"*.pth"']
    inclus = []
    if args.gif:
        exlus.remove('"*.gif"')
    if args.png:
        exlus.remove('"*.png"')
    if args.param:
        inclus.append('"params.pkl"')
        exlus.remove('"*.pkl"')

    if not args.include_all:
        for exc in exlus:
            cmd += ' --exclude ' + exc
        
        for inc in inclus:
            cmd += ' --include ' + inc
    
    print(cmd)
    # exit()
    subprocess.call(cmd, shell=True)


def main():
    parser = argparse.ArgumentParser(description='Process some integers.')
    parser.add_argument('log_dir', type=str, help='S3 Log dir')
    parser.add_argument('-b', '--bucket', type=str, default='chester-softgym', help='S3 Bucket')
    parser.add_argument('--param', type=int, default=0, help='Exclude')
    parser.add_argument('--gif', type=int, default=0, help='Exclude')
    parser.add_argument('--png', type=int, default=0, help='Exclude')
    parser.add_argument('--include_all', type=int, default=1, help='pull all data')

    args = parser.parse_args()
    s3_log_dir = "rllab/experiments/" + args.log_dir
    local_dir = os.path.join('./data', 'corl_s3_data', args.log_dir)
    if not os.path.exists(local_dir):
        os.makedirs(local_dir)
    aws_sync(args.bucket, s3_log_dir, local_dir, args)


if __name__ == "__main__":
    main()



# Add variants to finished experiments
import argparse
import os
import json
from pydoc import locate
import config


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('exp_folder', type=str, help='Root of the experiment folder to walk through')
    parser.add_argument('key', type=str, help='Name of the additional key')
    parser.add_argument('value', help='Value of the additional key')
    parser.add_argument('value_type', default='str', type=str, help='Type of the additional key')
    parser.add_argument('remote', nargs='?', default=None, type=str, ) # Optional

    args = parser.parse_args()
    exp_paths = [x[0] for x in os.walk(args.exp_folder, followlinks=True)]

    value_type = locate(args.value_type)
    if value_type == bool:
        value = args.value in ['1', 'True', 'true']
    else:
        value = value_type(args.value)

    for exp_path in exp_paths:
        try:
            variant_path = os.path.join(exp_path, "variant.json")
            # Modify locally
            with open(variant_path, 'r') as f:
                vv = json.load(f)
            if args.key in vv:
                print('Warning: key already in variants. {} = {}. Setting it to {}'.format(args.key, vv[args.key], value))

            vv[args.key] = value
            with open(variant_path, 'w') as f:
                json.dump(vv, f, indent=2, sort_keys=True)
            print('{} modified'.format(variant_path))

            # Upload it to remote
            if args.remote is not None:
                p = variant_path.rstrip('/').split('/')
                sub_exp_name, exp_name = p[-2], p[-3]

                remote_dir = os.path.join(config.REMOTE_DIR[args.remote], 'data', 'local', exp_name, sub_exp_name, 'variant.json')
                os.system('scp {} {}:{}'.format(variant_path, args.remote, remote_dir))
        except IOError as e:
            print(e)


if __name__ == '__main__':
    main()



"""
Chester - ML experiment launcher for local, SLURM, SSH, and EC2 environments.

Usage:
    from chester.run_exp import run_experiment_lite, VariantGenerator

    vg = VariantGenerator()
    vg.add('learning_rate', [0.001, 0.01])

    for variant in vg.variants():
        run_experiment_lite(
            stub_method_call=my_train_function,
            variant=variant,
            mode='local',
            exp_prefix='my_experiment',
        )
"""

__version__ = "0.2.0"
__author__ = "Chester Authors"

# Lazy imports to avoid requiring optional dependencies (hydra) at import time
# Users should import directly: from chester.run_exp import run_experiment_lite
__all__ = ["__version__"]


def __getattr__(name):
    """Lazy import for convenience access."""
    if name == "run_experiment_lite":
        from chester.run_exp import run_experiment_lite
        return run_experiment_lite
    elif name == "VariantGenerator":
        from chester.run_exp import VariantGenerator
        return VariantGenerator
    raise AttributeError(f"module 'chester' has no attribute '{name}'")



from chester import config_ec2 as config
from io import StringIO
import base64
import os
import os.path as osp
import subprocess
import hashlib
import datetime
import dateutil
import re

now = datetime.datetime.now(dateutil.tz.tzlocal())
timestamp = now.strftime('%Y_%m_%d_%H_%M_%S')


def dedent(s):
    lines = [l.strip() for l in s.split('\n')]
    return '\n'.join(lines)


def upload_file_to_s3(script_content):
    import tempfile
    import uuid
    f = tempfile.NamedTemporaryFile(delete=False)
    f.write(script_content.encode())
    f.close()
    remote_path = os.path.join(
        config.AWS_CODE_SYNC_S3_PATH, "oversize_bash_scripts", str(uuid.uuid4()))
    subprocess.check_call(["aws", "s3", "cp", f.name, remote_path])
    os.unlink(f.name)
    return remote_path


S3_CODE_PATH = None


def s3_sync_code(config, dry=False):
    global S3_CODE_PATH
    if S3_CODE_PATH is not None:
        return S3_CODE_PATH
    base = config.AWS_CODE_SYNC_S3_PATH
    has_git = True

    if config.FAST_CODE_SYNC:
        try:
            current_commit = subprocess.check_output(
                ["git", "rev-parse", "HEAD"]).strip().decode("utf-8")
        except subprocess.CalledProcessError as _:
            print("Warning: failed to execute git commands")
            current_commit = None
        file_name = str(timestamp) + "_" + hashlib.sha224(
            subprocess.check_output(["pwd"]) + str(current_commit).encode() + str(timestamp).encode()
        ).hexdigest() + ".tar.gz"

        file_path = "/tmp/" + file_name

        tar_cmd = ["tar", "-zcvf", file_path, "-C", config.PROJECT_PATH]
        for pattern in config.FAST_CODE_SYNC_IGNORES:
            tar_cmd += ["--exclude", pattern]
        tar_cmd += ["-h", "."]

        remote_path = "%s/%s" % (base, file_name)

        upload_cmd = ["aws", "s3", "cp", file_path, remote_path]

        mujoco_key_cmd = [
            "aws", "s3", "sync", config.MUJOCO_KEY_PATH, "{}/.mujoco/".format(base)]

        print(" ".join(tar_cmd))
        print(" ".join(upload_cmd))
        print(" ".join(mujoco_key_cmd))

        if not dry:
            subprocess.check_call(tar_cmd)
            subprocess.check_call(upload_cmd)
            try:
                subprocess.check_call(mujoco_key_cmd)
            except Exception as e:
                print(e)

        S3_CODE_PATH = remote_path
        return remote_path
    else:
        try:
            current_commit = subprocess.check_output(
                ["git", "rev-parse", "HEAD"]).strip().decode("utf-8")
            clean_state = len(
                subprocess.check_output(["git", "status", "--porcelain"])) == 0
        except subprocess.CalledProcessError as _:
            print("Warning: failed to execute git commands")
            has_git = False
        dir_hash = base64.b64encode(subprocess.check_output(["pwd"])).decode("utf-8")
        code_path = "%s_%s" % (
            dir_hash,
            (current_commit if clean_state else "%s_dirty_%s" % (current_commit, timestamp)) if
            has_git else timestamp
        )
        full_path = "%s/%s" % (base, code_path)
        cache_path = "%s/%s" % (base, dir_hash)
        cache_cmds = ["aws", "s3", "cp", "--recursive"] + \
                     flatten(["--exclude", "%s" % pattern] for pattern in config.CODE_SYNC_IGNORES) + \
                     [cache_path, full_path]
        cmds = ["aws", "s3", "cp", "--recursive"] + \
               flatten(["--exclude", "%s" % pattern] for pattern in config.CODE_SYNC_IGNORES) + \
               [".", full_path]
        caching_cmds = ["aws", "s3", "cp", "--recursive"] + \
                       flatten(["--exclude", "%s" % pattern] for pattern in config.CODE_SYNC_IGNORES) + \
                       [full_path, cache_path]
        mujoco_key_cmd = [
            "aws", "s3", "sync", config.MUJOCO_KEY_PATH, "{}/.mujoco/".format(base)]
        print(cache_cmds, cmds, caching_cmds, mujoco_key_cmd)
        if not dry:
            subprocess.check_call(cache_cmds)
            subprocess.check_call(cmds)
            subprocess.check_call(caching_cmds)
            try:
                subprocess.check_call(mujoco_key_cmd)
            except Exception:
                print('Unable to sync mujoco keys!')
        S3_CODE_PATH = full_path
        return full_path


_find_unsafe = re.compile(r'[a-zA-Z0-9_^@%+=:,./-]').search


def _shellquote(s):
    """Return a shell-escaped version of the string *s*."""
    if not s:
        return "''"

    if _find_unsafe(s) is None:
        return s

    # use single quotes, and put single quotes into double quotes
    # the string $'b is then quoted as '$'"'"'b'

    return "'" + s.replace("'", "'\"'\"'") + "'"


def _to_param_val(v):
    if v is None:
        return ""
    elif isinstance(v, list):
        return " ".join(map(_shellquote, list(map(str, v))))
    else:
        return _shellquote(str(v))


def to_local_command(params, python_command="python", script=osp.join(config.PROJECT_PATH, 'scripts/run_experiment.py'), use_gpu=False):
    command = python_command + " " + script
    pre_commands = params.pop("pre_commands", None)
    post_commands = params.pop("post_commands", None)
    if post_commands is not None:
        print("Not executing the post_commands: ", post_commands)

    for k, v in params.items():
        if isinstance(v, dict):
            for nk, nv in v.items():
                if str(nk) == "_name":
                    command += "  --%s %s" % (k, _to_param_val(nv))
                else:
                    command += \
                        "  --%s_%s %s" % (k, nk, _to_param_val(nv))
        else:
            command += "  --%s %s" % (k, _to_param_val(v))
    for pre_command in reversed(pre_commands):
        command = pre_command + " && " + command
    return command


def launch_ec2(params_list, exp_prefix, docker_image, code_full_path,
               python_command="python",
               script='scripts/run_experiment.py',
               aws_config=None, dry=False, terminate_machine=True, use_gpu=False, sync_s3_pkl=False,
               sync_s3_png=False,
               sync_s3_log=False,
               sync_s3_html=False,
               sync_s3_mp4=False,
               sync_s3_gif=False,
               sync_s3_pth=False,
               sync_s3_txt=False,
               sync_log_on_termination=True,
               periodic_sync=True, periodic_sync_interval=15):
    if len(params_list) == 0:
        return

    default_config = dict(
        image_id=config.AWS_IMAGE_ID,
        instance_type=config.AWS_INSTANCE_TYPE,
        key_name=config.AWS_KEY_NAME,
        spot=config.AWS_SPOT,
        spot_price=config.AWS_SPOT_PRICE,
        iam_instance_profile_name=config.AWS_IAM_INSTANCE_PROFILE_NAME,
        security_groups=config.AWS_SECURITY_GROUPS,
        security_group_ids=config.AWS_SECURITY_GROUP_IDS,
        network_interfaces=config.AWS_NETWORK_INTERFACES,
        instance_interruption_behavior='terminate',  # TODO
    )

    if aws_config is None:
        aws_config = dict()
    aws_config = dict(default_config, **aws_config)

    sio = StringIO()
    sio.write("#!/bin/bash\n")
    sio.write("{\n")
    sio.write("""
        die() { status=$1; shift; echo "FATAL: $*"; exit $status; }
    """)
    sio.write("""
        EC2_INSTANCE_ID="`wget -q -O - http://169.254.169.254/latest/meta-data/instance-id`"
    """)
    # sio.write("""service docker start""")
    # sio.write("""docker --config /home/ubuntu/.docker pull {docker_image}""".format(docker_image=docker_image))
    sio.write("""
        export PATH=/home/ubuntu/bin:/home/ubuntu/.local/bin:$PATH
    """)
    sio.write("""
        export PATH=/home/ubuntu/miniconda3/bin:/usr/local/cuda/bin:$PATH
    """)
    sio.write("""
        export LD_LIBRARY_PATH=$LD_LIBRARY_PATH:/root/.mujoco/mujoco200/bin
    """)
    sio.write("""
        echo $PATH
    """)
    sio.write("""
        export AWS_DEFAULT_REGION={aws_region}
    """.format(aws_region=config.AWS_BUCKET_REGION_NAME))  # add AWS_BUCKET_REGION_NAME=us-east-1 in your config.py
    sio.write("""
        pip install --upgrade --user awscli
    """)

    sio.write("""
        aws ec2 create-tags --resources $EC2_INSTANCE_ID --tags Key=Name,Value={exp_name} --region {aws_region}
    """.format(exp_name=params_list[0].get("exp_name"), aws_region=config.AWS_REGION_NAME))
    if config.LABEL:
        sio.write("""
            aws ec2 create-tags --resources $EC2_INSTANCE_ID --tags Key=owner,Value={label} --region {aws_region}
        """.format(label=config.LABEL, aws_region=config.AWS_REGION_NAME))
    sio.write("""
        aws ec2 create-tags --resources $EC2_INSTANCE_ID --tags Key=exp_prefix,Value={exp_prefix} --region {aws_region}
    """.format(exp_prefix=exp_prefix, aws_region=config.AWS_REGION_NAME))

    if config.FAST_CODE_SYNC:
        sio.write("""
            aws s3 cp {code_full_path} /tmp/chester_code.tar.gz
        """.format(code_full_path=code_full_path, local_code_path=config.CODE_DIR))
        sio.write("""
            mkdir -p {local_code_path}
        """.format(code_full_path=code_full_path, local_code_path=config.CODE_DIR,
                   aws_region=config.AWS_REGION_NAME))
        sio.write("""
            tar -zxvf /tmp/chester_code.tar.gz -C {local_code_path}
        """.format(code_full_path=code_full_path, local_code_path=config.CODE_DIR,
                   aws_region=config.AWS_REGION_NAME))
    else:
        sio.write("""
            aws s3 cp --recursive {code_full_path} {local_code_path}
        """.format(code_full_path=code_full_path, local_code_path=config.CODE_DIR))
    s3_mujoco_key_path = config.AWS_CODE_SYNC_S3_PATH + '/.mujoco/'
    sio.write("""
        aws s3 cp --recursive {} {}
    """.format(s3_mujoco_key_path, config.MUJOCO_KEY_PATH))
    sio.write("""
        cd {local_code_path}
    """.format(local_code_path=config.CODE_DIR))

    for params in params_list:
        log_dir = params.get("log_dir")
        remote_log_dir = params.pop("remote_log_dir")
        env = params.pop("env", None)

        sio.write("""
            aws ec2 create-tags --resources $EC2_INSTANCE_ID --tags Key=Name,Value={exp_name} --region {aws_region}
        """.format(exp_name=params.get("exp_name"), aws_region=config.AWS_REGION_NAME))
        sio.write("""
            mkdir -p {log_dir}
        """.format(log_dir=log_dir))
        if periodic_sync:
            include_png = " --include '*.png' " if sync_s3_png else " "
            include_pkl = " --include '*.pkl' " if sync_s3_pkl else " "
            include_log = " --include '*.log' " if sync_s3_log else " "
            include_html = " --include '*.html' " if sync_s3_html else " "
            include_mp4 = " --include '*.mp4' " if sync_s3_mp4 else " "
            include_gif = " --include '*.gif' " if sync_s3_gif else " "
            include_pth = " --include '*.pth' " if sync_s3_pth else " "
            include_txt = " --include '*.txt' " if sync_s3_txt else " "
            sio.write("""
                while /bin/true; do
                    aws s3 sync --exclude '*' {include_png} {include_pkl} {include_log} {include_html} {include_mp4} {include_gif} {include_pth} {include_txt} --include '*.csv' --include '*.json' {log_dir} {remote_log_dir}
                    sleep {periodic_sync_interval}
                done & echo sync initiated""".format(include_png=include_png, include_pkl=include_pkl,
                                                     include_log=include_log, include_html=include_html,
                                                     include_mp4=include_mp4, include_gif=include_gif,
                                                     include_pth=include_pth, include_txt=include_txt,
                                                     log_dir=log_dir, remote_log_dir=remote_log_dir,
                                                     periodic_sync_interval=periodic_sync_interval))
            if sync_log_on_termination:
                sio.write("""
                    while /bin/true; do
                        if [ -z $(curl -Is http://169.254.169.254/latest/meta-data/spot/termination-time | head -1 | grep 404 | cut -d \  -f 2) ]
                          then
                            logger "Running shutdown hook."
                            aws s3 cp /home/ubuntu/user_data.log {remote_log_dir}/stdout.log
                            aws s3 cp --recursive {log_dir} {remote_log_dir}
                            break
                          else
                            # Spot instance not yet marked for termination.
                            sleep 5
                        fi
                    done & echo log sync initiated
                """.format(log_dir=log_dir, remote_log_dir=remote_log_dir))
        sio.write("""{command}""".format(command=to_local_command(params, python_command=python_command, script=script, use_gpu=use_gpu)))
        sio.write("""
        aws s3 cp --recursive {log_dir} {remote_log_dir}
         """.format(log_dir=log_dir, remote_log_dir=remote_log_dir))
        sio.write("""
        aws s3 cp /home/ubuntu/user_data.log {remote_log_dir}/stdout.log
        """.format(remote_log_dir=remote_log_dir))

    if terminate_machine:
        sio.write("""
            EC2_INSTANCE_ID="`wget -q -O - http://169.254.169.254/latest/meta-data/instance-id || die \"wget instance-id has failed: $?\"`"
            aws ec2 terminate-instances --instance-ids $EC2_INSTANCE_ID --region {aws_region}
        """.format(aws_region=config.AWS_REGION_NAME))
    sio.write("} >> /home/ubuntu/user_data.log 2>&1\n")

    full_script = dedent(sio.getvalue())

    import boto3
    import botocore
    if aws_config["spot"]:
        ec2 = boto3.client(
            "ec2",
            region_name=config.AWS_REGION_NAME,
            aws_access_key_id=config.AWS_ACCESS_KEY,
            aws_secret_access_key=config.AWS_ACCESS_SECRET,
        )
    else:
        ec2 = boto3.resource(
            "ec2",
            region_name=config.AWS_REGION_NAME,
            aws_access_key_id=config.AWS_ACCESS_KEY,
            aws_secret_access_key=config.AWS_ACCESS_SECRET,
        )

    print("len_full_script", len(full_script))
    if len(full_script) > 16384 or len(base64.b64encode(full_script.encode()).decode("utf-8")) > 16384:
        # Script too long; need to upload script to s3 first.
        # We're being conservative here since the actual limit is 16384 bytes
        s3_path = upload_file_to_s3(full_script)
        sio = StringIO()
        sio.write("#!/bin/bash\n")
        sio.write("""
        aws s3 cp {s3_path} /home/ubuntu/remote_script.sh --region {aws_region} && \\
        chmod +x /home/ubuntu/remote_script.sh && \\
        bash /home/ubuntu/remote_script.sh
        """.format(s3_path=s3_path, aws_region=config.AWS_REGION_NAME))
        user_data = dedent(sio.getvalue())
    else:
        user_data = full_script
    print(full_script)
    with open("/tmp/full_script", "w") as f:
        f.write(full_script)

    instance_args = dict(
        ImageId=aws_config["image_id"],
        KeyName=aws_config["key_name"],
        UserData=user_data,
        InstanceType=aws_config["instance_type"],
        EbsOptimized=config.EBS_OPTIMIZED,
        SecurityGroups=aws_config["security_groups"],
        SecurityGroupIds=aws_config["security_group_ids"],
        NetworkInterfaces=aws_config["network_interfaces"],
        IamInstanceProfile=dict(
            Name=aws_config["iam_instance_profile_name"],
        ),
        **config.AWS_EXTRA_CONFIGS,
    )

    if len(instance_args["NetworkInterfaces"]) > 0:
        # disable_security_group = query_yes_no(
        #     "Cannot provide both network interfaces and security groups info. Do you want to disable security group settings?",
        #     default="yes",
        # )
        disable_security_group = True
        if disable_security_group:
            instance_args.pop("SecurityGroups")
            instance_args.pop("SecurityGroupIds")

    if aws_config.get("placement", None) is not None:
        instance_args["Placement"] = aws_config["placement"]
    if not aws_config["spot"]:
        instance_args["MinCount"] = 1
        instance_args["MaxCount"] = 1
    print("************************************************************")
    print(instance_args["UserData"])
    print("************************************************************")
    if aws_config["spot"]:
        instance_args["UserData"] = base64.b64encode(instance_args["UserData"].encode()).decode("utf-8")
        spot_args = dict(
            DryRun=dry,
            InstanceCount=1,
            LaunchSpecification=instance_args,
            SpotPrice=aws_config["spot_price"],
            # ClientToken=params_list[0]["exp_name"],
        )
        import pprint
        pprint.pprint(spot_args)
        if not dry:
            response = ec2.request_spot_instances(**spot_args)
            print(response)
            spot_request_id = response['SpotInstanceRequests'][
                0]['SpotInstanceRequestId']
            for _ in range(10):
                try:
                    ec2.create_tags(
                        Resources=[spot_request_id],
                        Tags=[
                            {'Key': 'Name', 'Value': params_list[0]["exp_name"]}
                        ],
                    )
                    break
                except botocore.exceptions.ClientError:
                    continue
    else:
        import pprint
        pprint.pprint(instance_args)
        ec2.q(
            DryRun=dry,
            **instance_args
        )



"""
Chester Configuration Module

This module provides configuration for chester via YAML files.
Configuration is loaded from chester.yaml in the parent project directory.

Search order for chester.yaml:
1. CHESTER_CONFIG_PATH environment variable
2. Parent directories from current working directory (up to git root)
3. Fall back to defaults for local-only usage

Usage:
    from chester import config
    print(config.PROJECT_PATH)
    print(config.REMOTE_DIR['gl'])
"""

import os
import warnings
from pathlib import Path
from typing import Any, Dict, List, Optional

# Lazy-loaded configuration
_config: Optional[Dict[str, Any]] = None
_config_path: Optional[Path] = None


def _find_config_file() -> Optional[Path]:
    """
    Search for chester.yaml configuration file.

    Search order:
    1. CHESTER_CONFIG_PATH environment variable
    2. Parent directories from cwd (stops at git root or filesystem root)

    Returns:
        Path to chester.yaml if found, None otherwise
    """
    # Check environment variable first
    env_path = os.environ.get('CHESTER_CONFIG_PATH')
    if env_path:
        path = Path(env_path)
        if path.exists():
            return path
        warnings.warn(f"CHESTER_CONFIG_PATH set to {env_path} but file not found")

    # Search parent directories from cwd
    current = Path.cwd().resolve()
    while current != current.parent:
        config_path = current / 'chester.yaml'
        if config_path.exists():
            return config_path
        # Stop at git root to avoid searching too far
        if (current / '.git').exists():
            break
        current = current.parent

    return None


def _get_defaults() -> Dict[str, Any]:
    """
    Get default configuration for local-only usage.

    These defaults allow chester to work without a config file
    for basic local execution.
    """
    cwd = os.getcwd()
    return {
        'project_path': cwd,
        'log_dir': os.path.join(cwd, 'data'),
        'host_address': {'local': ''},
        'ssh_hosts': [],
        'remote_dir': {'local': cwd},
        'remote_log_dir': {'local': os.path.join(cwd, 'data')},
        'remote_header': {},
        'simg_path': {},
        'cuda_module': {},
        'modules': {},
        'remote_mount_option': {},
        'autobot_nodelist': [],
        'gpu_state_dir': '',
        'chester_queue_dir': '',
        'chester_scheduler_log_dir': '',
        # Package manager config
        'package_manager': 'uv',  # 'uv' or 'conda'
        'conda_env': None,  # Required if package_manager is 'conda'
        'conda_command': 'conda',  # 'conda' or 'mamba'
        'sync_on_launch': True,  # Whether to run uv sync / conda env update
        # Rsync patterns (optional, falls back to file-based approach)
        'rsync_include': [],
        'rsync_exclude': [],
    }


def _resolve_paths(config: Dict[str, Any], config_dir: Path) -> Dict[str, Any]:
    """
    Resolve relative paths in configuration.

    - project_path: defaults to config file directory if not specified
    - log_dir: resolved relative to project_path if not absolute
    """
    # Resolve project_path
    if 'project_path' not in config or not config['project_path']:
        config['project_path'] = str(config_dir)
    else:
        config['project_path'] = os.path.abspath(config['project_path'])

    project_path = config['project_path']

    # Resolve log_dir relative to project_path
    log_dir = config.get('log_dir', 'data')
    if not os.path.isabs(log_dir):
        config['log_dir'] = os.path.join(project_path, log_dir)
    else:
        config['log_dir'] = log_dir

    return config


def _merge_with_defaults(config: Dict[str, Any]) -> Dict[str, Any]:
    """
    Merge loaded config with defaults for missing keys.
    """
    defaults = _get_defaults()
    for key, default_value in defaults.items():
        if key not in config:
            config[key] = default_value
        elif isinstance(default_value, dict) and isinstance(config.get(key), dict):
            # Merge dict values (e.g., host_address)
            merged = default_value.copy()
            merged.update(config[key])
            config[key] = merged
    return config


def _load_config() -> Dict[str, Any]:
    """
    Load configuration from chester.yaml with lazy initialization.

    Returns:
        Configuration dictionary
    """
    global _config, _config_path

    if _config is not None:
        return _config

    _config_path = _find_config_file()

    if _config_path is not None:
        try:
            import yaml
            with open(_config_path) as f:
                _config = yaml.safe_load(f) or {}
            _config = _resolve_paths(_config, _config_path.parent)
            _config = _merge_with_defaults(_config)
        except ImportError:
            warnings.warn(
                "PyYAML not installed. Install with: pip install pyyaml\n"
                "Falling back to default configuration."
            )
            _config = _get_defaults()
        except Exception as e:
            warnings.warn(f"Error loading {_config_path}: {e}\nFalling back to defaults.")
            _config = _get_defaults()
    else:
        # No config file found - use defaults
        _config = _get_defaults()

    return _config


def get_config_path() -> Optional[Path]:
    """Get the path to the loaded configuration file, if any."""
    _load_config()  # Ensure config is loaded
    return _config_path


def reload_config() -> Dict[str, Any]:
    """
    Force reload of configuration.

    Useful for testing or when config file has changed.
    """
    global _config, _config_path
    _config = None
    _config_path = None
    return _load_config()


# Mapping from UPPER_CASE attribute names to yaml keys
_ATTR_MAPPING = {
    'PROJECT_PATH': 'project_path',
    'LOG_DIR': 'log_dir',
    'HOST_ADDRESS': 'host_address',
    'SSH_HOSTS': 'ssh_hosts',
    'REMOTE_DIR': 'remote_dir',
    'REMOTE_LOG_DIR': 'remote_log_dir',
    'REMOTE_HEADER': 'remote_header',
    'SIMG_PATH': 'simg_path',
    'CUDA_MODULE': 'cuda_module',
    'MODULES': 'modules',
    'REMOTE_MOUNT_OPTION': 'remote_mount_option',
    'AUTOBOT_NODELIST': 'autobot_nodelist',
    'GPU_STATE_DIR': 'gpu_state_dir',
    'CHESTER_QUEUE_DIR': 'chester_queue_dir',
    'CHESTER_CHEDULER_LOG_DIR': 'chester_scheduler_log_dir',  # Note: typo preserved for compatibility
    # Package manager config
    'PACKAGE_MANAGER': 'package_manager',
    'CONDA_ENV': 'conda_env',
    'CONDA_COMMAND': 'conda_command',
    'SYNC_ON_LAUNCH': 'sync_on_launch',
    # Rsync patterns
    'RSYNC_INCLUDE': 'rsync_include',
    'RSYNC_EXCLUDE': 'rsync_exclude',
}


def __getattr__(name: str) -> Any:
    """
    Enable config.VARIABLE_NAME access pattern.

    This allows backward-compatible access like:
        config.PROJECT_PATH
        config.REMOTE_DIR['gl']
    """
    if name in _ATTR_MAPPING:
        cfg = _load_config()
        yaml_key = _ATTR_MAPPING[name]
        value = cfg.get(yaml_key)

        # Handle empty string defaults for path configs
        if value is None:
            if yaml_key in ('project_path', 'log_dir', 'gpu_state_dir',
                           'chester_queue_dir', 'chester_scheduler_log_dir'):
                return ''
            return {} if yaml_key.endswith('_dir') or yaml_key in (
                'host_address', 'remote_dir', 'remote_log_dir', 'remote_header',
                'simg_path', 'cuda_module', 'modules', 'remote_mount_option'
            ) else []

        return value

    # Check if it's a module-level function
    if name in ('get_config_path', 'reload_config', '_load_config'):
        return globals()[name]

    raise AttributeError(f"module 'chester.config' has no attribute '{name}'")


def __dir__() -> List[str]:
    """List available attributes for tab completion."""
    return list(_ATTR_MAPPING.keys()) + ['get_config_path', 'reload_config']


# For debugging: print config when run directly
if __name__ == "__main__":
    cfg = _load_config()
    print(f"Config file: {_config_path}")
    print(f"PROJECT_PATH: {cfg.get('project_path')}")
    print(f"LOG_DIR: {cfg.get('log_dir')}")
    print(f"HOST_ADDRESS: {cfg.get('host_address')}")
    print(f"REMOTE_DIR: {cfg.get('remote_dir')}")



#!/usr/bin/env python
"""
Local tests for chester - no remote access required.

Run with: uv run python tests/test_local.py
"""
import os
import tempfile


def test_variant_generator():
    """Test VariantGenerator creates correct parameter combinations."""
    print("=" * 60)
    print("TEST: VariantGenerator")
    print("=" * 60)

    from chester.run_exp import VariantGenerator

    vg = VariantGenerator()
    vg.add('learning_rate', [0.001, 0.01])
    vg.add('batch_size', [32, 64])
    vg.add('hidden_dim', lambda batch_size: [batch_size * 2])  # Dependency

    variants = vg.variants()

    print(f"Generated {len(variants)} variants:")
    for i, v in enumerate(variants):
        print(f"  {i}: lr={v.learning_rate}, bs={v.batch_size}, hd={v.hidden_dim}")

    # Verify count: 2 * 2 = 4 variants
    assert len(variants) == 4, f"Expected 4 variants, got {len(variants)}"

    # Verify dependency works
    for v in variants:
        assert v.hidden_dim == v.batch_size * 2, "Dependency not working"

    # Verify first/last markers
    assert variants[0].get('chester_first_variant', False), "Missing first_variant marker"
    assert variants[-1].get('chester_last_variant', False), "Missing last_variant marker"

    print("PASSED\n")


def test_config_loading():
    """Test YAML-based config loading."""
    print("=" * 60)
    print("TEST: Config Loading")
    print("=" * 60)

    from chester import config

    # Test basic attributes exist
    print(f"PROJECT_PATH: {config.PROJECT_PATH}")
    print(f"LOG_DIR: {config.LOG_DIR}")
    print(f"HOST_ADDRESS: {config.HOST_ADDRESS}")

    assert isinstance(config.PROJECT_PATH, str), "PROJECT_PATH should be a string"
    assert isinstance(config.LOG_DIR, str), "LOG_DIR should be a string"
    assert isinstance(config.HOST_ADDRESS, dict), "HOST_ADDRESS should be a dict"
    assert 'local' in config.HOST_ADDRESS, "HOST_ADDRESS should have 'local' key"

    print("PASSED\n")


def test_local_command_generation():
    """Test local command generation."""
    print("=" * 60)
    print("TEST: Local Command Generation")
    print("=" * 60)

    from chester.slurm import to_local_command

    params = {
        'log_dir': '/tmp/test',
        'exp_name': 'test_exp',
        'args_data': 'base64encodeddata',
    }

    cmd = to_local_command(
        params=params,
        python_command='python',
        script='chester/run_exp_worker.py',
        env={'MY_VAR': 'value'}
    )

    print(f"Generated command:\n  {cmd}\n")

    assert 'python' in cmd, "Missing python command"
    assert 'run_exp_worker.py' in cmd, "Missing script"
    assert 'log_dir' in cmd, "Missing log_dir param"

    print("PASSED\n")


def test_slurm_command_generation():
    """Test SLURM command generation (dry run - no submission)."""
    print("=" * 60)
    print("TEST: SLURM Command Generation")
    print("=" * 60)

    from chester.slurm import to_slurm_command

    params = {
        'log_dir': '/tmp/test',
        'exp_name': 'test_exp',
        'args_data': 'base64encodeddata',
    }

    header = """#!/usr/bin/env bash
#SBATCH --nodes=1
#SBATCH --partition=gpu
#SBATCH --gres=gpu:1
#SBATCH --time=1:00:00"""

    cmd_list = to_slurm_command(
        params=params,
        use_gpu=True,
        modules=['singularity'],
        cuda_module='cuda/11.8',
        header=header,
        python_command='python',
        script='train.py',
        use_singularity=False,
        simg_dir=None,
        remote_dir='/home/user/project',
        mount_options='',
        set_egl_gpu=False,
        env={},
    )

    full_cmd = "\n".join(cmd_list)
    print(f"Generated SLURM script:\n{full_cmd[:500]}...\n")

    assert '#SBATCH' in full_cmd, "Missing SBATCH header"
    assert 'module load' in full_cmd, "Missing module load"
    assert 'train.py' in full_cmd, "Missing script"

    print("PASSED\n")


def test_logger():
    """Test logger functionality."""
    print("=" * 60)
    print("TEST: Logger")
    print("=" * 60)

    from chester import logger

    with tempfile.TemporaryDirectory() as tmpdir:
        # Configure logger with CSV output
        logger.configure(dir=tmpdir, format_strs=['csv', 'stdout'])

        # Log some values
        logger.logkv('epoch', 1)
        logger.logkv('loss', 0.5)
        logger.logkv('accuracy', 0.85)
        logger.dumpkvs()

        logger.logkv('epoch', 2)
        logger.logkv('loss', 0.3)
        logger.logkv('accuracy', 0.92)
        logger.dumpkvs()

        # Check CSV was created
        csv_path = os.path.join(tmpdir, 'progress.csv')
        assert os.path.exists(csv_path), f"CSV file not created at {csv_path}"

        with open(csv_path) as f:
            content = f.read()
            print(f"CSV content:\n{content}")
            assert 'epoch' in content, "Missing epoch column"
            assert 'loss' in content, "Missing loss column"

        logger.reset()

    print("PASSED\n")


def test_ssh_command_generation():
    """Test SSH command generation for non-SLURM hosts."""
    print("=" * 60)
    print("TEST: SSH Command Generation")
    print("=" * 60)

    from chester.slurm import to_ssh_command

    params = {
        'log_dir': '/tmp/test',
        'exp_name': 'test_exp',
        'args_data': 'base64encodeddata',
    }

    cmd_list = to_ssh_command(
        params=params,
        python_command='python',
        remote_dir='/home/user/project',
        script='train.py',
        env={'CUDA_VISIBLE_DEVICES': '0'},
    )

    full_cmd = "\n".join(cmd_list)
    print(f"Generated SSH script:\n{full_cmd}\n")

    assert 'cd /home/user/project' in full_cmd, "Missing cd command"
    assert 'train.py' in full_cmd, "Missing script"
    assert '.done' in full_cmd, "Missing .done marker creation"

    print("PASSED\n")


def main():
    print("\n" + "=" * 60)
    print("CHESTER LOCAL TESTS")
    print("=" * 60 + "\n")

    tests = [
        test_variant_generator,
        test_config_loading,
        test_local_command_generation,
        test_slurm_command_generation,
        test_logger,
        test_ssh_command_generation,
    ]

    passed = 0
    failed = 0
    skipped = 0

    for test in tests:
        try:
            test()
            passed += 1
        except AssertionError as e:
            print(f"FAILED: {e}\n")
            failed += 1
        except Exception as e:
            if "SKIPPED" in str(e) or "not found" in str(e):
                skipped += 1
            else:
                print(f"ERROR: {e}\n")
                import traceback
                traceback.print_exc()
                failed += 1

    print("=" * 60)
    print(f"RESULTS: {passed} passed, {failed} failed, {skipped} skipped")
    print("=" * 60)

    return 0 if failed == 0 else 1


if __name__ == '__main__':
    import sys
    sys.exit(main())



import os
import os.path as osp
import re
from subprocess import run
from tempfile import NamedTemporaryFile
from chester import config
from chester.hydra_utils import to_hydra_command

# TODO remove the singularity part

slurm_dir = './'


def get_package_manager_setup_commands(sync=True, source_prepare=True):
    """
    Generate shell commands for package manager setup based on config.

    Args:
        sync: Whether to sync/update the environment
        source_prepare: Whether to source prepare.sh for custom project setup

    Returns list of shell commands to set up the environment.
    """
    pkg_manager = config.PACKAGE_MANAGER
    sync_on_launch = config.SYNC_ON_LAUNCH if sync else False
    commands = []

    if pkg_manager == 'uv':
        # Add common uv installation paths to PATH
        commands.append('export PATH="$HOME/.local/bin:$HOME/.cargo/bin:$PATH"')
        # Auto-install uv if not present
        commands.append(
            'if ! command -v uv &> /dev/null; then '
            'echo "[chester] Installing uv..."; '
            'curl -LsSf https://astral.sh/uv/install.sh | sh; '
            'export PATH="$HOME/.local/bin:$PATH"; '
            'fi'
        )
        # Set CUDA_HOME for packages that need it
        commands.append(
            'export CUDA_HOME=$(dirname $(dirname $(which nvcc 2>/dev/null || echo /usr/local/cuda/bin/nvcc)))'
        )
        commands.append('echo "[chester] CUDA_HOME=$CUDA_HOME"')
        # Set fake versions for setuptools-scm packages
        commands.append('export SETUPTOOLS_SCM_PRETEND_VERSION_FOR_NVIDIA_CUROBO=0.7.0')
        # Set MAX_JOBS for parallel CUDA compilation
        commands.append('export MAX_JOBS=16')
        # Install ninja globally for faster builds
        commands.append('uv tool install ninja 2>/dev/null || true')
        # Sync environment
        if sync_on_launch:
            commands.append('echo "[chester] Running uv sync..."')
            commands.append('uv sync')

    elif pkg_manager == 'conda':
        conda_cmd = config.CONDA_COMMAND or 'conda'
        conda_env = config.CONDA_ENV
        if not conda_env:
            raise ValueError("chester config: conda_env is required when package_manager is 'conda'")

        # Source bashrc to get conda in PATH
        commands.append('source ~/.bashrc')
        # Activate conda environment
        commands.append(f'{conda_cmd} activate {conda_env}')
        # Set CUDA_HOME
        commands.append(
            'export CUDA_HOME=$(dirname $(dirname $(which nvcc 2>/dev/null || echo /usr/local/cuda/bin/nvcc)))'
        )
        commands.append('echo "[chester] CUDA_HOME=$CUDA_HOME"')
        # Sync environment (update from environment.yml if exists)
        if sync_on_launch:
            commands.append(
                f'if [ -f environment.yml ]; then '
                f'echo "[chester] Running {conda_cmd} env update..."; '
                f'{conda_cmd} env update -n {conda_env} -f environment.yml --prune; '
                f'fi'
            )
    else:
        raise ValueError(f"chester config: unknown package_manager '{pkg_manager}' (expected 'uv' or 'conda')")

    # Source prepare.sh for custom project setup (if exists)
    if source_prepare:
        commands.append('if [ -f ./prepare.sh ]; then . ./prepare.sh; fi')

    return commands


def get_python_command(base_command="python"):
    """
    Get the python command wrapped for the configured package manager.

    Args:
        base_command: Base python command (e.g., "python", "srun python")

    Returns:
        Wrapped command string (e.g., "uv run python", "conda run -n env python")
    """
    pkg_manager = config.PACKAGE_MANAGER

    if pkg_manager == 'uv':
        # For uv, prefix with "uv run"
        return f'uv run {base_command}'
    elif pkg_manager == 'conda':
        # For conda, the environment is already activated, no wrapper needed
        return base_command
    else:
        return base_command


def slurm_run_scripts(scripts):
    """this is another function that those _sub files should call. this actually execute files"""
    # TODO support running multiple scripts

    assert isinstance(scripts, str)

    os.chdir(slurm_dir)

    # make sure it will run.
    assert scripts.startswith('#!/usr/bin/env bash\n')
    file_temp = NamedTemporaryFile(delete=False)
    file_temp.write(scripts.encode('utf-8'))
    file_temp.close()
    run(['sbatch', file_temp.name], check=True)
    os.remove(file_temp.name)


_find_unsafe = re.compile(r'[a-zA-Z0-9_^@%+=:,./-]').search


def _shellquote(s):
    """Return a shell-escaped version of the string *s*."""
    if not s:
        return "''"

    if _find_unsafe(s) is None:
        return s

    # use single quotes, and put single quotes into double quotes
    # the string $'b is then quoted as '$'"'"'b'

    return "'" + s.replace("'", "'\"'\"'") + "'"


def _to_param_val(v):
    if v is None:
        return ""
    elif isinstance(v, list):
        return " ".join(map(_shellquote, list(map(str, v))))
    else:
        return _shellquote(str(v))

def to_local_command(
    params, 
    python_command="python", 
    script=None,
    env={}):
    command = python_command + " " + script

    for k, v in env.items():
        command = ("%s=%s " % (k, v)) + command
    pre_commands = params.pop("pre_commands", None)
    post_commands = params.pop("post_commands", None)
    if pre_commands is not None or post_commands is not None:
        print("Not executing the pre_commands: ", pre_commands, ", nor post_commands: ", post_commands)

    for k, v in params.items():
        if isinstance(v, dict):
            for nk, nv in v.items():
                if str(nk) == "_name":
                    command += "  --%s %s" % (k, _to_param_val(nv))
                else:
                    command += \
                        "  --%s_%s %s" % (k, nk, _to_param_val(nv))
        else:
            command += "  --%s %s" % (k, _to_param_val(v))
    return command

def to_slurm_command(params, header, python_command="srun python", remote_dir='~/',
                     script=None,  # Defer config access to runtime
                     simg_dir=None, use_gpu=False, modules=None, cuda_module=None, use_singularity=True,
                     mount_options=None, compile_script=None, wait_compile=None, set_egl_gpu=False,
                     env=None, hydra_enabled=False, hydra_flags=None, sync_env=None):
    """
    Transfer the commands to the format that can be run by slurm.

    Package manager (uv/conda) is determined by chester.yaml config.

    Args:
        params: Dictionary of parameters
        header: SLURM header with #SBATCH directives
        python_command: Base python command (will be wrapped based on package_manager)
        remote_dir: Remote directory
        script: Script to run
        simg_dir: Singularity image directory
        use_gpu: Whether to use GPU
        modules: List of modules to load
        cuda_module: CUDA module to load
        use_singularity: Whether to use singularity container
        mount_options: Singularity mount options
        compile_script: Optional compile script
        wait_compile: Time to wait after compile
        set_egl_gpu: Whether to set EGL_GPU env var
        env: Environment variables
        hydra_enabled: Whether to use Hydra
        hydra_flags: Hydra flags
        sync_env: Override for sync_on_launch config (None = use config)
    """
    # Resolve default script path at runtime (not import time)
    if script is None:
        script = osp.join(config.PROJECT_PATH, 'scripts/run_experiment.py')

    command_list = list()
    command_list.append(header)
    sing_commands = list()

    command_list.append('set -x')  # echo commands to stdout
    command_list.append('set -u')  # throw an error if unset variable referenced
    command_list.append('set -e')  # exit on errors
    command_list.append('srun hostname')
    command_list.append('cd {}'.format(remote_dir))
    for remote_module in modules:
        command_list.append('module load ' + remote_module)
    if use_gpu:
        assert cuda_module is not None
        command_list.append('module load ' + cuda_module)

    # Log into singularity shell
    if use_singularity:
        if mount_options is not None:
            options = '-B ' + mount_options
        else:
            options = ''
        sing_prefix = 'singularity exec {} {} {} /bin/bash -c'.format(options, '--nv' if use_gpu else '', simg_dir)
    else:
        sing_prefix = '/bin/bash -c'

    # Package manager setup (replaces hardcoded prepare.sh sourcing)
    do_sync = sync_env if sync_env is not None else config.SYNC_ON_LAUNCH
    setup_commands = get_package_manager_setup_commands(sync=do_sync)
    sing_commands.extend(setup_commands)

    # NCCL settings for distributed training
    sing_commands.append("export NCCL_DEBUG=INFO")
    sing_commands.append("export PYTHONFAULTHANDLER=1")
    sing_commands.append("export NCCL_SOCKET_IFNAME=^docker0,lo")

    if set_egl_gpu:
        sing_commands.append('export EGL_GPU=$SLURM_JOB_GRES')
        sing_commands.append('echo $EGL_GPU')
    if compile_script is not None:
        sing_commands.append(compile_script)
    if wait_compile is not None:
        sing_commands.append('sleep ' + str(int(wait_compile)))

    pre_commands = params.pop("pre_commands", None)
    post_commands = params.pop("post_commands", None)
    if pre_commands is not None:
        command_list.extend(pre_commands)

    # Wrap python command for package manager
    wrapped_python = get_python_command(python_command)

    if hydra_enabled:
        command = to_hydra_command(params, wrapped_python, script, hydra_flags, env)
    else:
        command = to_local_command(params, wrapped_python, script, env)
    sing_commands.append(command)

    # Add .done marker file on successful completion (for auto_pull to detect)
    assert 'log_dir' in params, "chester auto_pull: log_dir must be in params"
    log_dir = params['log_dir']
    sing_commands.append(f'touch {log_dir}/.done')

    all_sing_cmds = ' && '.join(sing_commands)
    command_list.append(sing_prefix + ' \'{}\''.format(all_sing_cmds))
    if post_commands is not None:
        command_list.extend(post_commands)
    return command_list

def to_ssh_command(params, python_command="python", remote_dir='./',
                   script='main.py', env=None, hydra_enabled=False, hydra_flags=None,
                   sync_env=None):
    """
    Generate a bash script for SSH-based remote execution (no SLURM).

    This creates a simpler script than to_slurm_command, suitable for
    running on remote machines via SSH + nohup.

    Package manager (uv/conda) is determined by chester.yaml config.

    Args:
        params: Dictionary of parameters to pass to the script
        python_command: Base python command (will be wrapped based on package_manager config)
        remote_dir: Remote directory to cd into
        script: Script to run
        env: Environment variables to set
        hydra_enabled: Whether to use Hydra command format
        hydra_flags: Hydra flags if hydra_enabled
        sync_env: Override for sync_on_launch config (None = use config)
    """
    command_list = []

    # Bash header
    command_list.append('#!/usr/bin/env bash')
    command_list.append('set -x')  # echo commands to stdout
    command_list.append('set -u')  # throw an error if unset variable referenced
    command_list.append('set -e')  # exit on errors

    # Change to remote directory
    command_list.append(f'cd {remote_dir}')
    # Add project root to PYTHONPATH for local modules
    command_list.append('export PYTHONPATH="$PWD:${PYTHONPATH:-}"')

    # Package manager setup (uv or conda based on config)
    do_sync = sync_env if sync_env is not None else config.SYNC_ON_LAUNCH
    setup_commands = get_package_manager_setup_commands(sync=do_sync)
    command_list.extend(setup_commands)

    # Wrap python command for package manager
    wrapped_python = get_python_command(python_command)

    # Generate the main command
    if hydra_enabled:
        command = to_hydra_command(params, wrapped_python, script, hydra_flags, env)
    else:
        command = to_local_command(params, wrapped_python, script, env if env else {})
    command_list.append(command)

    # Add .done marker file on successful completion (for auto_pull to detect)
    assert 'log_dir' in params, "chester auto_pull: log_dir must be in params"
    log_dir = params['log_dir']
    command_list.append(f'touch {log_dir}/.done')

    return command_list


# if __name__ == '__main__':
#     slurm_run_scripts(header)



import os
from pathlib import Path
import re
import subprocess
import base64
import os.path as osp
import pickle as pickle
import cloudpickle
import numpy as np
import inspect
import collections
import glob
import sys
import time
import datetime
import dateutil.tz
import json
from chester import config, config_ec2
from chester.hydra_utils import to_hydra_command, run_hydra_command
from chester.slurm import to_slurm_command, to_local_command, to_ssh_command
from chester.utils_s3 import launch_ec2, s3_sync_code


# Auto-pull manifest management
_auto_pull_manifest_path = None
_auto_pull_jobs = []


def _map_local_to_remote_log_dir(local_log_dir: str, mode: str) -> str:
    """
    Map a local log directory to its remote equivalent.

    The user specifies a local log_dir, and chester automatically maps it to
    the equivalent path on remote based on project roots.

    Example:
        Local PROJECT_PATH: /home/user/project
        Local log_dir: /home/user/project/data/train/exp1
        Remote remote_dir: /home/remote/project
        -> Remote log_dir: /home/remote/project/data/train/exp1

    Args:
        local_log_dir: Local log directory (absolute or relative to PROJECT_PATH)
        mode: Execution mode (e.g., 'armfranka', 'gl')

    Returns:
        Remote log directory path

    Raises:
        ValueError: If local_log_dir is not within PROJECT_PATH
    """
    project_path = config.PROJECT_PATH
    remote_dir = config.REMOTE_DIR.get(mode)

    if not remote_dir:
        raise ValueError(f"No remote_dir configured for mode '{mode}'")

    # Resolve local_log_dir to absolute path
    if not os.path.isabs(local_log_dir):
        local_log_dir = os.path.join(project_path, local_log_dir)
    local_log_dir = os.path.normpath(local_log_dir)

    # Ensure local_log_dir is within PROJECT_PATH
    project_path_normalized = os.path.normpath(project_path)
    if not local_log_dir.startswith(project_path_normalized + os.sep) and local_log_dir != project_path_normalized:
        raise ValueError(
            f"log_dir must be within PROJECT_PATH for remote sync.\n"
            f"  log_dir: {local_log_dir}\n"
            f"  PROJECT_PATH: {project_path_normalized}\n"
            f"Log directory must be a subdirectory of the project root."
        )

    # Compute relative path from PROJECT_PATH
    relative_path = os.path.relpath(local_log_dir, project_path_normalized)

    # Map to remote
    remote_log_dir = os.path.join(remote_dir, relative_path)
    return remote_log_dir


def _init_auto_pull_manifest(exp_prefix: str, mode: str):
    """Initialize the manifest file path for this batch of experiments."""
    global _auto_pull_manifest_path, _auto_pull_jobs
    manifest_dir = os.path.join(config.LOG_DIR, '.chester_manifests')
    os.makedirs(manifest_dir, exist_ok=True)
    timestamp = datetime.datetime.now().strftime('%m%d_%H%M%S')
    _auto_pull_manifest_path = os.path.join(manifest_dir, f'{exp_prefix}_{mode}_{timestamp}.json')
    _auto_pull_jobs = []


def _add_job_to_manifest(host: str, remote_log_dir: str, local_log_dir: str, exp_name: str):
    """Add a job to the auto-pull manifest."""
    global _auto_pull_jobs
    _auto_pull_jobs.append({
        'host': host,
        'remote_log_dir': remote_log_dir,
        'local_log_dir': local_log_dir,
        'exp_name': exp_name,
        'status': 'pending',
        'submitted_at': datetime.datetime.now().isoformat()
    })


def _save_and_spawn_auto_pull(dry: bool = False, poll_interval: int = 60):
    """Save the manifest and spawn the auto-pull poller."""
    global _auto_pull_manifest_path, _auto_pull_jobs

    if not _auto_pull_jobs:
        print('[chester] No jobs to track for auto-pull')
        return

    assert _auto_pull_manifest_path is not None, "chester auto_pull: manifest path not initialized"

    # Save manifest
    with open(_auto_pull_manifest_path, 'w') as f:
        json.dump(_auto_pull_jobs, f, indent=2)
    print(f'[chester] Saved auto-pull manifest: {_auto_pull_manifest_path}')
    print(f'[chester] Tracking {len(_auto_pull_jobs)} jobs for auto-pull')

    if dry:
        print('[chester] Dry run - not spawning auto-pull poller')
        return

    # Spawn background poller
    chester_dir = os.path.dirname(os.path.abspath(__file__))
    auto_pull_script = os.path.join(chester_dir, 'auto_pull.py')
    log_dir = os.path.join(config.LOG_DIR, '.chester_manifests')
    log_file = _auto_pull_manifest_path.replace('.json', '.log')

    cmd = (f'nohup python {auto_pull_script} '
           f'--manifest {_auto_pull_manifest_path} '
           f'--poll-interval {poll_interval} '
           f'> {log_file} 2>&1 &')

    print(f'[chester] Spawning auto-pull poller: {cmd}')
    os.system(cmd)
    print(f'[chester] Auto-pull log: {log_file}')


def monitor_processes(active_processes, max_processes=2, sleep_time=1):
    """
    Monitor the number of running processes and wait if maximum is reached.
    
    Args:
        process_list: List of subprocess.Popen objects
        max_processes: Maximum number of concurrent processes
        sleep_time: Time to sleep between checks in seconds
        
    Returns:
        Updated list with only running processes
    """
    # Wait if we've reached the maximum number of processes
    while len(active_processes) >= max_processes:
        time.sleep(sleep_time)
        # Update the list of active processes
        active_processes = [p for p in active_processes if p.poll() is None]
    
    return active_processes

def query_yes_no(question, default="yes"):
    """Ask a yes/no question via raw_input() and return their answer.

    "question" is a string that is presented to the user.
    "default" is the presumed answer if the user just hits <Enter>.
        It must be "yes" (the default), "no" or None (meaning
        an answer is required of the user).

    The "answer" return value is True for "yes" or False for "no".
    """
    valid = {"yes": True, "y": True, "ye": True,
             "no": False, "n": False}
    if default is None:
        prompt = " [y/n] "
    elif default == "yes":
        prompt = " [Y/n] "
    elif default == "no":
        prompt = " [y/N] "
    else:
        raise ValueError("invalid default answer: '%s'" % default)

    while True:
        sys.stdout.write(question + prompt)
        choice = input().lower()
        if default is not None and choice == '':
            return valid[default]
        elif choice in valid:
            return valid[choice]
        else:
            sys.stdout.write("Please respond with 'yes' or 'no' "
                             "(or 'y' or 'n').\n")


_find_unsafe = re.compile(r'[a-zA-Z0-9_^@%+=:,./-]').search


def _shellquote(s):
    """Return a shell-escaped version of the string *s*."""
    if not s:
        return "''"

    if _find_unsafe(s) is None:
        return s

    # use single quotes, and put single quotes into double quotes
    # the string $'b is then quoted as '$'"'"'b'

    return "'" + s.replace("'", "'\"'\"'") + "'"


def _to_param_val(v):
    if v is None:
        return ""
    elif isinstance(v, list):
        return " ".join(map(_shellquote, list(map(str, v))))
    else:
        return _shellquote(str(v))




class AttrDict(dict):
    def __init__(self, *args, **kwargs):
        super(AttrDict, self).__init__(*args, **kwargs)
        self.__dict__ = self


class VariantDict(AttrDict):
    def __init__(self, d, hidden_keys):
        super(VariantDict, self).__init__(d)
        self._hidden_keys = hidden_keys

    def dump(self):
        return {k: v for k, v in self.items() if k not in self._hidden_keys}


class VariantGenerator(dict):
    """
    Usage:

    vg = VariantGenerator()
    vg.add("param1", [1, 2, 3])
    vg.add("param2", ['x', 'y'])
    vg.variants() => # all combinations of [1,2,3] x ['x','y']

    Supports noncyclic dependency among parameters:
    vg = VariantGenerator()
    vg.add("param1", [1, 2, 3])
    vg.add("param2", lambda param1: [param1+1, param1+2])
    vg.variants() => # ..
    """

    def __init__(self):
        self._variants = []
        self._populate_variants()
        self._hidden_keys = []
        for k, vs, cfg in self._variants:
            if cfg.get("hide", False):
                self._hidden_keys.append(k)

    @property
    def size(self):
        return len(self.variants())

    def __getitem__(self, item):
        for param in self.variations():
            if param[0] == item:
                return param[1]

    def add(self, key, vals, **kwargs):
        self._variants.append((key, vals, kwargs))

    def update(self, key, vals, **kwargs):
        for i, (k, _, _) in enumerate(self._variants):
            if k == key:
                self._variants[i] = (k, vals, kwargs)
                return
        self.add(key, vals, kwargs)

    def _populate_variants(self):
        methods = inspect.getmembers(
            self.__class__, predicate=lambda x: inspect.isfunction(x) or inspect.ismethod(x))
        methods = [x[1].__get__(self, self.__class__)
                   for x in methods if getattr(x[1], '__is_variant', False)]
        for m in methods:
            self.add(m.__name__, m, **getattr(m, "__variant_config", dict()))

    def variations(self):
        ret = []
        for key, vals, _ in self._variants:
            if not isinstance(vals, list):
                continue
            if len(vals) > 1:
                ret.append(key)
        return ret

    def variants(self, randomized=False):
        ret = list(self.ivariants())
        if randomized:
            np.random.shuffle(ret)
        # ret = list(map(self.variant_dict, ret))
        ret = list(map(AttrDict, ret))
        ret[0]['chester_first_variant'] = True
        ret[-1]['chester_last_variant'] = True
        return ret

    def variant_dict(self, variant):
        return VariantDict(variant, self._hidden_keys)

    def to_name_suffix(self, variant):
        suffix = []
        for k, vs, cfg in self._variants:
            if not cfg.get("hide", False):
                suffix.append(k + "_" + str(variant[k]))
        return "_".join(suffix)

    def ivariants(self):
        dependencies = list()
        for key, vals, cfg in self._variants:
            if cfg.get("hide", False):
                continue  # Skip hidden keys entirely
            if hasattr(vals, "__call__"):
                args = inspect.getfullargspec(vals).args
                if hasattr(vals, 'im_self') or hasattr(vals, "__self__"):
                    # remove the first 'self' parameter
                    args = args[1:]
                dependencies.append((key, set(args)))
            else:
                dependencies.append((key, set()))
        sorted_keys = []
        # topo sort all nodes
        while len(sorted_keys) < len(self._variants):
            # get all nodes with zero in-degree
            free_nodes = [k for k, v in dependencies if len(v) == 0]
            if len(free_nodes) == 0:
                error_msg = "Invalid parameter dependency: \n"
                for k, v in dependencies:
                    if len(v) > 0:
                        error_msg += k + " depends on " + " & ".join(v) + "\n"
                raise ValueError(error_msg)
            dependencies = [(k, v)
                            for k, v in dependencies if k not in free_nodes]
            # remove the free nodes from the remaining dependencies
            for _, v in dependencies:
                v.difference_update(free_nodes)
            sorted_keys += free_nodes
        return self._ivariants_sorted(sorted_keys)

    def _ivariants_sorted(self, sorted_keys):
        if len(sorted_keys) == 0:
            yield dict()
        else:
            first_keys = sorted_keys[:-1]
            first_variants = self._ivariants_sorted(first_keys)
            last_key = sorted_keys[-1]
            last_vals = [v for k, v, _ in self._variants if k == last_key][0]
            if hasattr(last_vals, "__call__"):
                last_val_keys = inspect.getfullargspec(last_vals).args
                if hasattr(last_vals, 'im_self') or hasattr(last_vals, '__self__'):
                    last_val_keys = last_val_keys[1:]
            else:
                last_val_keys = None
            for variant in first_variants:
                if hasattr(last_vals, "__call__"):
                    last_variants = last_vals(
                        **{k: variant[k] for k in last_val_keys})
                    for last_choice in last_variants:
                        yield AttrDict(variant, **{last_key: last_choice})
                else:
                    for last_choice in last_vals:
                        yield AttrDict(variant, **{last_key: last_choice})


def variant(*args, **kwargs):
    def _variant(fn):
        fn.__is_variant = True
        fn.__variant_config = kwargs
        return fn

    if len(args) == 1 and isinstance(args[0], collections.Callable):
        return _variant(args[0])
    return _variant


def rsync_code(remote_host, remote_dir):
    """
    Sync project code to remote host.

    Syncs from PROJECT_PATH to remote_dir.
    Requires rsync_include and rsync_exclude lists in chester.yaml.
    """
    project_path = config.PROJECT_PATH
    print(f'Ready to rsync code: {project_path} -> {remote_host}:{remote_dir}')

    yaml_include = config.RSYNC_INCLUDE
    yaml_exclude = config.RSYNC_EXCLUDE

    if not yaml_include and not yaml_exclude:
        raise ValueError("rsync_include and rsync_exclude must be defined in chester.yaml")

    include_args = ' '.join(f"--include='{p}'" for p in yaml_include)
    exclude_args = ' '.join(f"--exclude='{p}'" for p in yaml_exclude)
    cmd = f"rsync -avzh --delete {include_args} {exclude_args} {project_path}/ {remote_host}:{remote_dir}"
    print(cmd)
    os.system(cmd)


exp_count = -2
sub_process_popens = []
now = datetime.datetime.now(dateutil.tz.tzlocal())
timestamp = now.strftime('%m_%d_%H_%M')
remote_confirmed = False


def run_experiment_lite(
        stub_method_call=None,
        batch_tasks=None,
        exp_prefix="experiment",
        exp_name=None,
        log_dir=None,
        sub_dir='train',
        script=None,  # TODO: change this before making pip package
        python_command="python",
        mode="local",
        use_gpu=False,
        dry=False,
        env={},
        variant=None,
        variations=[],
        use_cloudpickle=True,
        pre_commands=None,
        print_command=True,
        launch_with_subprocess=True,
        wait_subprocess=True,
        max_num_processes=10,
        compile_script=None,
        wait_compile=None,
        use_singularity=False,
        hydra_enabled=False,  # New parameter for Hydra support
        hydra_flags=None,  # Additional hydra flags like multirun
        auto_pull=False,  # Enable automatic result pulling from remote
        auto_pull_interval=60,  # Poll interval in seconds for auto-pull
        sync_env=None,  # Override sync_on_launch config (None = use config)
        **kwargs):
    """
    Serialize the stubbed method call and run the experiment using the specified mode.
    :param stub_method_call: A stubbed method call.
    :param script: The name of the entrance point python script
    :param mode: Where & how to run the experiment. Can be ['local', 'singularity', 'seuss', 'psc']
    :param dry: Whether to do a dry-run, which only prints the commands without executing them.
    :param exp_prefix: Name prefix for the experiments
    :param env: extra environment variables
    :param kwargs: All other parameters will be passed directly to the entrance python script.
    :param variant: If provided, should be a dictionary of parameters
    :param hydra_enabled: If True, will use Hydra command line override format instead of chester format
    :param hydra_flags: Optional dictionary of hydra flags to add to the command (e.g., {'multirun': True})
    """
    last_variant = variant.pop('chester_last_variant', False)
    first_variant = variant.pop('chester_first_variant', False)
    host = config.HOST_ADDRESS[mode]
    local_chester_queue_dir = config.LOG_DIR + '/queues'
    remote_sub_dir = os.path.join(config.REMOTE_LOG_DIR[mode], sub_dir)
    remote_batch_dir = os.path.join(remote_sub_dir, exp_prefix)
    local_batch_dir = os.path.join(config.LOG_DIR, sub_dir, exp_prefix)

    if first_variant:
        os.system(f'mkdir -p {local_chester_queue_dir}')
    if mode == 'singularity':
        mode = 'local_singularity'

    assert stub_method_call is not None or batch_tasks is not None or script is not None, \
        "Must provide at least either stub_method_call or batch_tasks or script"
    if script is None:
        script = '-m chester.run_exp_worker'  # Use module syntax for installed package
    
    if batch_tasks is None:
        batch_tasks = [
            dict(
                kwargs,
                pre_commands=pre_commands,
                stub_method_call=stub_method_call,
                exp_name=exp_name,
                log_dir=log_dir,
                env=env,
                variant=variant,
                use_cloudpickle=use_cloudpickle
            )
        ]

    global exp_count
    global remote_confirmed
    global sub_process_popens
    sub_process_popens = monitor_processes(sub_process_popens, max_num_processes)

    for task in batch_tasks:
        call = task.pop("stub_method_call")
        if use_cloudpickle:
            data = base64.b64encode(cloudpickle.dumps(call)).decode("utf-8")
        else:
            data = base64.b64encode(pickle.dumps(call)).decode("utf-8")
        task["args_data"] = data
        exp_count += 1
        
        if task.get("exp_name", None) is None:
            exp_name = exp_prefix
            for v in variations:
                print(v)
                key_name = v.split('.')[-1]
                if isinstance(variant[v], (list, tuple)):
                    continue
                if isinstance(variant[v], str):
                    exp_name += '_{}'.format(variant[v].split('/')[-1])
                elif isinstance(variant[v], bool):
                    if variant[v]:
                        exp_name += '_{}'.format(key_name)
                elif variant[v] is not None:  # int or float
                    exp_name += '_{}_{:g}'.format(key_name, variant[v])
            ind = len(glob.glob(os.path.join(local_batch_dir, '[0-9]*_*')))
            if exp_count == -1:
                exp_count = ind + 1
            task["exp_name"] = "{}_{}".format(exp_count, exp_name)
            print('exp name ', task["exp_name"])
        # Handle log_dir: user specifies local path, chester maps to remote
        # Store original local log_dir for auto-pull, compute remote log_dir for execution
        local_log_dir = task.get("log_dir", None)
        if local_log_dir is None:
            # Default: data/{sub_dir}/{exp_prefix}/{exp_name}
            local_log_dir = os.path.join(config.LOG_DIR, sub_dir, exp_prefix, task["exp_name"])
        elif not os.path.isabs(local_log_dir):
            # Relative path - resolve relative to PROJECT_PATH
            local_log_dir = os.path.join(config.PROJECT_PATH, local_log_dir)
        local_log_dir = os.path.normpath(local_log_dir)
        task['_local_log_dir'] = local_log_dir  # Store for auto-pull

        # For remote modes, map local to remote; for local mode, use as-is
        if mode in ["local", "local_singularity"]:
            task['log_dir'] = local_log_dir
        else:
            task['log_dir'] = _map_local_to_remote_log_dir(local_log_dir, mode)

        if task.get("variant", None) is not None:
            variant = task.pop("variant")
            if "exp_name" not in variant:
                variant["exp_name"] = task["exp_name"]
                # variant["group_name"] = exp_prefix
            task["variant_data"] = base64.b64encode(pickle.dumps(variant)).decode("utf-8")
        elif "variant" in task:
            del task["variant"]
        task["env"] = task.get("env", dict()) or dict()
        local_exp_dir = os.path.join(local_batch_dir, task["exp_name"])

    if mode not in ["local", "local_singularity", "ec2"] and not remote_confirmed and not dry:
        remote_confirmed = query_yes_no(
            "Running in (non-dry) mode %s. Confirm?" % mode)
        if not remote_confirmed:
            sys.exit(1)

    if mode in ["local"]:
        for task in batch_tasks:
            env = task.pop("env", None)
            task.pop('_local_log_dir', None)  # Pop internal field, not for CLI
            # Generate command based on whether hydra is enabled or not
            if hydra_enabled:
                command = to_hydra_command(
                    params=task,
                    python_command=python_command,
                    script=script,
                    hydra_flags=hydra_flags,
                    env=env
                )
            else:
                command = to_local_command(
                    task,
                    python_command=python_command,
                    script=script,
                    env=env
                )
                
            if print_command:
                print(command)
            if dry:
                return
            
            if env is None:
                env = dict()
            if launch_with_subprocess:
                try:
                    if wait_subprocess:
                        subprocess.call(command, shell=True, env=dict(os.environ, **env))
                        popen_obj = None
                    else:
                        popen_obj = subprocess.Popen(command, shell=True, env=dict(os.environ, **env))
                    sub_process_popens.append(popen_obj)
                except Exception as e:
                    print(e)
                    if isinstance(e, KeyboardInterrupt):
                        raise
            else: # for debug, not need to catch subprocess
                assert hydra_enabled, "hydra_enabled must be True when launch_with_subprocess is False"
                run_hydra_command(command, task["log_dir"], stub_method_call)
                popen_obj = None

            return popen_obj
    elif mode == 'local_singularity':
        for task in batch_tasks:
            env = task.pop("env", None)
            task.pop('_local_log_dir', None)  # Pop internal field, not for CLI
            command = to_local_command(
                task,
                python_command=python_command,
                script=osp.join(config.PROJECT_PATH, script)
            )
            if print_command:
                print(command)
            if dry:
                return
            try:
                if env is None:
                    env = dict()
                # TODO add argument for specifying container
                singularity_header = f'singularity exec {config.SIMG_PATH[mode]}'
                command = singularity_header + ' ' + command
                subprocess.call(
                    command, shell=True, env=dict(os.environ, **env))
                popen_obj = None
            except Exception as e:
                print(e)
                if isinstance(e, KeyboardInterrupt):
                    raise
            return popen_obj
    elif mode in ['gl', 'seuss', 'psc', 'satori']:
        for task in batch_tasks:
            remote_dir = config.REMOTE_DIR[mode]
            simg_dir = config.SIMG_PATH[mode]
            if first_variant and not dry:
                rsync_code(remote_host=host, remote_dir=remote_dir)
            if first_variant and auto_pull:
                _init_auto_pull_manifest(exp_prefix, mode)
            remote_log_dir = task['log_dir']
            local_log_dir = task.pop('_local_log_dir')  # Pop internal field, not for CLI
            header = config.REMOTE_HEADER[mode]
            header = header + "\n#SBATCH -o " + os.path.join(remote_log_dir, 'slurm.out') + " # STDOUT"
            header = header + "\n#SBATCH -e " + os.path.join(remote_log_dir, 'slurm.err') + " # STDERR\n"
            gpus = str(variant.get("gpus", 1))
            header = header.replace("$gpus", gpus)
            python_command = "srun " + python_command
            command_list = to_slurm_command(
                task,
                use_gpu=use_gpu,
                modules=config.MODULES[mode],
                cuda_module=config.CUDA_MODULE[mode],
                header=header,
                python_command=python_command,
                script=script,
                use_singularity=use_singularity,
                simg_dir=simg_dir,
                remote_dir=remote_dir,
                mount_options=config.REMOTE_MOUNT_OPTION[mode],
                compile_script=compile_script,
                wait_compile=wait_compile,
                hydra_enabled=hydra_enabled,
                hydra_flags=hydra_flags,
                set_egl_gpu=False,
                env=env
            )
            if print_command:
                print("; ".join(command_list))
            command = "\n".join(command_list)
            os.system(f'mkdir -p {local_exp_dir}')
            local_script_name = os.path.join(local_exp_dir, "slurm_launch")
            with open(local_script_name, 'w') as f:
                f.write(command)

            # Add job to auto-pull manifest
            if auto_pull:
                _add_job_to_manifest(
                    host=host,
                    remote_log_dir=remote_log_dir,
                    local_log_dir=local_log_dir,
                    exp_name=task['exp_name']
                )

            if last_variant:
                print('Remote sub dir: ', remote_sub_dir)
                os.system('scp -r {f1} {host}:{f2}'.format(f1=local_batch_dir, f2=remote_sub_dir, host=mode))
                os.system(f'rm -rf {local_batch_dir}')
                print('Ready to execute the scheduler')
                remote_cmd = (f'cd {remote_dir} && . ./prepare.sh && '
                              f'python chester/scheduler/remote_slurm_launcher.py {remote_batch_dir} {int(dry)}')
                cmd = "ssh  {host} \'{cmd} \'".format(host=host,
                                                      cmd=remote_cmd
                                                      )
                print('Submit to slurm ', cmd)
                os.system(cmd)  # Launch

                # Save manifest and spawn auto-pull poller
                if auto_pull:
                    _save_and_spawn_auto_pull(dry=dry, poll_interval=auto_pull_interval)
            # Cleanup
    elif mode in ['autobot']:
        for task in batch_tasks:
            # TODO check remote directory
            remote_dir = config.REMOTE_DIR[mode]
            simg_dir = config.SIMG_PATH[mode]

            # query_yes_no('Confirm: Syncing code to {}:{}'.format(mode, remote_dir))
            if first_variant:
                rsync_code(remote_host=mode, remote_dir=remote_dir)
            if first_variant and auto_pull:
                _init_auto_pull_manifest(exp_prefix, mode)
            remote_log_dir = task['log_dir']
            local_log_dir = task.pop('_local_log_dir')  # Pop internal field, not for CLI
            remote_script_name = os.path.join(remote_log_dir, task['exp_name'])
            local_script_name = os.path.join(local_exp_dir, task['exp_name'])
            header = '#CHESTERNODE ' + ','.join(config.AUTOBOT_NODELIST)
            header = header + "\n#CHESTEROUT " + os.path.join(remote_log_dir, 'slurm.out')
            header = header + "\n#CHESTERERR " + os.path.join(remote_log_dir, 'slurm.err')
            header = header + "\n#CHESTERSCRIPT " + remote_script_name
            if simg_dir.find('$') == -1:
                simg_dir = osp.join(remote_dir, simg_dir)
            command_list = to_slurm_command(
                task,
                use_gpu=use_gpu,
                modules=config.MODULES[mode],
                cuda_module=config.CUDA_MODULE[mode],
                header=header,
                python_command=python_command,
                use_singularity=use_singularity,
                script=osp.join(remote_dir, script),
                simg_dir=simg_dir,
                remote_dir=remote_dir,
                mount_options=config.REMOTE_MOUNT_OPTION[mode],
                compile_script=compile_script,
                wait_compile=wait_compile,
                set_egl_gpu=True,
            )
            if print_command:
                print("; ".join(command_list))
            command = "\n".join(command_list)
            script_name = './data/tmp/' + task['exp_name']
            scheduler_script_name = os.path.join(local_chester_queue_dir, task['exp_name'])
            with open(script_name, 'w') as f:
                f.write(command)
            os.system(f'cp {script_name} {local_script_name}')
            os.system(f'cp {script_name} {scheduler_script_name}')

            # Add job to auto-pull manifest
            if auto_pull:
                _add_job_to_manifest(
                    host=mode,
                    remote_log_dir=remote_log_dir,
                    local_log_dir=local_log_dir,
                    exp_name=task['exp_name']
                )

            # Cleanup
            os.remove(script_name)
            # Open scheduler if all jobs have been submitted
            # Remote end will only open another scheduler when there is not one running already
            # Redirect the output of the remote scheduler to the log file
            if last_variant and not dry:
                print('Syncing to remote')
                print('Remote batch dir: ', remote_batch_dir)
                os.system('scp -r {f1} {host}:{f2}'.format(f1=local_batch_dir, f2=remote_batch_dir, host=mode))
                os.system('scp -r {f1} {host}:{f2}'.format(f1=local_chester_queue_dir, f2=config.CHESTER_QUEUE_DIR,
                                                           host=mode))
                os.system(f'rm -rf {local_batch_dir}')
                os.system(f'rm -rf {local_chester_queue_dir}')
                # os.system("ssh {host} \'{cmd}\'".format(host=mode, cmd='mkdir -p ' + config.CHESTER_CHEDULER_LOG_DIR))
                t = datetime.datetime.now(dateutil.tz.tzlocal()).strftime('%m_%d_%H_%M')
                # log_file = os.path.join(config.CHESTER_CHEDULER_LOG_DIR, f'{t}.txt')
                log_file = os.path.join(config.CHESTER_CHEDULER_LOG_DIR, f'log_{t}.txt')
                print('Ready to execute the scheduler')
                cmd = "ssh  {host} \'{cmd} > {output}&\'".format(host=mode,
                                                                 cmd=f'cd {remote_dir} && . ./prepare.sh && nohup python chester/scheduler/remote_scheduler.py',
                                                                 output=log_file)
                if dry:
                    print(remote_script_name)
                    print(cmd)
                else:
                    print(cmd)
                    os.system(cmd)

                # Save manifest and spawn auto-pull poller
                if auto_pull:
                    _save_and_spawn_auto_pull(dry=dry, poll_interval=auto_pull_interval)

    elif mode in config.SSH_HOSTS:
        # SSH-based remote execution (no SLURM)
        # Package manager (uv/conda) determined by chester.yaml config
        for task in batch_tasks:
            remote_dir = config.REMOTE_DIR[mode]

            if first_variant and not dry:
                rsync_code(remote_host=host, remote_dir=remote_dir)
            if first_variant and auto_pull:
                _init_auto_pull_manifest(exp_prefix, mode)

            remote_log_dir = task['log_dir']
            local_log_dir = task.pop('_local_log_dir')  # Pop internal field, not for CLI

            # Pop env from task dict to pass separately to to_ssh_command
            task_env = task.pop("env", None)
            if task_env:
                # Merge with any env passed to run_experiment_lite
                merged_env = {**env, **task_env} if env else task_env
            else:
                merged_env = env

            # Generate SSH command script (python command wrapped by slurm.py based on config)
            command_list = to_ssh_command(
                task,
                python_command=python_command,
                remote_dir=remote_dir,
                script=script,
                env=merged_env,
                hydra_enabled=hydra_enabled,
                hydra_flags=hydra_flags,
                sync_env=sync_env,
            )

            if print_command:
                print("; ".join(command_list))

            command = "\n".join(command_list)

            # Save script locally
            os.system(f'mkdir -p {local_exp_dir}')
            local_script_name = os.path.join(local_exp_dir, "ssh_launch.sh")
            with open(local_script_name, 'w') as f:
                f.write(command)

            # Add job to auto-pull manifest
            if auto_pull:
                _add_job_to_manifest(
                    host=host,
                    remote_log_dir=remote_log_dir,
                    local_log_dir=local_log_dir,
                    exp_name=task['exp_name']
                )

            if not dry:
                # Copy script to remote
                remote_script_path = os.path.join(remote_log_dir, "ssh_launch.sh")
                os.system(f'ssh {host} "mkdir -p {remote_log_dir}"')
                os.system(f'scp {local_script_name} {host}:{remote_script_path}')

                # Execute via SSH with nohup
                stdout_log = os.path.join(remote_log_dir, 'stdout.log')
                stderr_log = os.path.join(remote_log_dir, 'stderr.log')
                ssh_cmd = (f'ssh {host} "cd {remote_dir} && '
                           f'nohup bash {remote_script_path} > {stdout_log} 2> {stderr_log} &"')

                print(f'[chester] Launching on {host}: {task["exp_name"]}')
                print(f'[chester] Remote log dir: {remote_log_dir}')
                os.system(ssh_cmd)

        # Save manifest and spawn auto-pull poller (after all jobs submitted)
        if last_variant and auto_pull:
            _save_and_spawn_auto_pull(dry=dry, poll_interval=auto_pull_interval)

    elif mode == 'ec2':
        # if docker_image is None:
        #     docker_image = config.DOCKER_IMAGE
        s3_code_path = s3_sync_code(config_ec2, dry=dry)
        for task in batch_tasks:
            task["remote_log_dir"] = osp.join(config_ec2.AWS_S3_PATH, exp_prefix.replace("_", "-"), task["exp_name"])
            if compile_script is None:
                task["pre_commands"] = [". ./prepare_ec2.sh", 'time ./compile.sh']
            else:
                task["pre_commands"] = [". ./prepare_ec2.sh", 'time ./' + compile_script]
        launch_ec2(batch_tasks,
                   exp_prefix=exp_prefix,
                   docker_image=None,  # Currently not using docker
                   python_command=python_command,
                   script=script,
                   aws_config=None,
                   dry=dry,
                   terminate_machine=True,
                   use_gpu=use_gpu,
                   code_full_path=s3_code_path,
                   sync_s3_pkl=True,
                   sync_s3_html=True,
                   sync_s3_png=True,
                   sync_s3_log=True,
                   sync_s3_gif=True,
                   sync_s3_mp4=True,
                   sync_s3_pth=True,
                   sync_s3_txt=True,
                   sync_log_on_termination=True,
                   periodic_sync=True,
                   periodic_sync_interval=15)



import sys
import argparse
import os.path as osp
import datetime
import dateutil.tz
import ast
import uuid
import pickle as pickle
import base64
import joblib
import sys
# sys.path.insert(0, '/home/zixuanhu/UMD')
from chester import config
# sys.path.insert(0, PROJECT_PATH)


def run_experiment(argv):
    default_log_dir = config.LOG_DIR
    now = datetime.datetime.now(dateutil.tz.tzlocal())

    # avoid name clashes when running distributed jobs
    rand_id = str(uuid.uuid4())[:5]
    timestamp = now.strftime('%Y_%m_%d_%H_%M_%S_%f_%Z')

    default_exp_name = 'experiment_%s_%s' % (timestamp, rand_id)
    parser = argparse.ArgumentParser()
    parser.add_argument('--n_parallel', type=int, default=1,
                        help='Number of parallel workers to perform rollouts. 0 => don\'t start any workers')
    parser.add_argument(
        '--exp_name', type=str, default=default_exp_name, help='Name of the experiment.')
    parser.add_argument('--log_dir', type=str, default=None,
                        help='Path to save the log and iteration snapshot.')
    parser.add_argument('--snapshot_mode', type=str, default='all',
                        help='Mode to save the snapshot. Can be either "all" '
                             '(all iterations will be saved), "last" (only '
                             'the last iteration will be saved), "gap" (every'
                             '`snapshot_gap` iterations are saved), or "none" '
                             '(do not save snapshots)')
    parser.add_argument('--snapshot_gap', type=int, default=1,
                        help='Gap between snapshot iterations.')
    parser.add_argument('--tabular_log_file', type=str, default='progress.csv',
                        help='Name of the tabular log file (in csv).')
    parser.add_argument('--text_log_file', type=str, default='debug.log',
                        help='Name of the text log file (in pure text).')
    parser.add_argument('--params_log_file', type=str, default='params.json',
                        help='Name of the parameter log file (in json).')
    parser.add_argument('--variant_log_file', type=str, default='variant.json',
                        help='Name of the variant log file (in json).')
    parser.add_argument('--resume_from', type=str, default=None,
                        help='Name of the pickle file to resume experiment from.')
    parser.add_argument('--plot', type=ast.literal_eval, default=False,
                        help='Whether to plot the iteration results')
    parser.add_argument('--log_tabular_only', type=ast.literal_eval, default=False,
                        help='Whether to only print the tabular log information (in a horizontal format)')
    parser.add_argument('--seed', type=int,
                        help='Random seed for numpy')
    parser.add_argument('--args_data', type=str,
                        help='Pickled data for stub objects')
    parser.add_argument('--variant_data', type=str,
                        help='Pickled data for variant configuration')
    parser.add_argument('--use_cloudpickle', type=ast.literal_eval, default=False)

    args = parser.parse_args(argv[1:])

    # if args.seed is not None:
    #     set_seed(args.seed)
    #
    # if args.n_parallel > 0:
    #     from rllab.sampler import parallel_sampler
    #     parallel_sampler.initialize(n_parallel=args.n_parallel)
    #     if args.seed is not None:
    #         parallel_sampler.set_seed(args.seed)
    #
    # if args.plot:
    #     from rllab.plotter import plotter
    #     plotter.init_worker()

    if args.log_dir is None:
        log_dir = osp.join(default_log_dir, args.exp_name)
    else:
        log_dir = args.log_dir
    # tabular_log_file = osp.join(log_dir, args.tabular_log_file)
    # text_log_file = osp.join(log_dir, args.text_log_file)
    # params_log_file = osp.join(log_dir, args.params_log_file)

    if args.variant_data is not None:
        variant_data = pickle.loads(base64.b64decode(args.variant_data))
        variant_log_file = osp.join(log_dir, args.variant_log_file)
        # logger.log_variant(variant_log_file, variant_data)
    else:
        variant_data = None

    # if not args.use_cloudpickle:
    #     logger.log_parameters_lite(params_log_file, args)
    #
    # logger.add_text_output(text_log_file)
    # logger.add_tabular_output(tabular_log_file)
    # prev_snapshot_dir = logger.get_snapshot_dir()
    # prev_mode = logger.get_snapshot_mode()
    # logger.set_snapshot_dir(log_dir)
    # logger.set_snapshot_mode(args.snapshot_mode)
    # logger.set_snapshot_gap(args.snapshot_gap)
    # logger.set_log_tabular_only(args.log_tabular_only)
    # logger.push_prefix("[%s] " % args.exp_name)

    if args.resume_from is not None:
        data = joblib.load(args.resume_from)
        assert 'algo' in data
        algo = data['algo']
        algo.train()
    else:
        # read from stdin
        if args.use_cloudpickle:
            import cloudpickle
            method_call = cloudpickle.loads(base64.b64decode(args.args_data))
            method_call(variant_data, log_dir, args.exp_name)
        else:
            assert False
            # data = pickle.loads(base64.b64decode(args.args_data))
            # maybe_iter = concretize(data)
            # if is_iterable(maybe_iter):
            #     for _ in maybe_iter:
            #         pass

            # logger.set_snapshot_mode(prev_mode)
            # logger.set_snapshot_dir(prev_snapshot_dir)
            # logger.remove_tabular_output(tabular_log_file)
            # logger.remove_text_output(text_log_file)
            # logger.pop_prefix()


if __name__ == "__main__":
    run_experiment(sys.argv)



#!/usr/bin/env python
"""
Chester Auto Pull - Background poller for automatic result synchronization.

This script polls remote hosts for .done marker files and automatically
pulls results back to the local machine when jobs complete.

Usage:
    # Run as background process (spawned by chester)
    python chester/auto_pull.py --manifest /path/to/manifest.json

    # Run manually to pull specific jobs
    python chester/auto_pull.py --manifest /path/to/manifest.json --once
"""

import os
import sys
import json
import time
import argparse
import subprocess
from pathlib import Path
from datetime import datetime

from chester import config


def check_done_marker(host: str, remote_log_dir: str) -> bool:
    """Check if .done marker exists on remote host via SSH."""
    done_file = os.path.join(remote_log_dir, '.done')
    cmd = f'ssh {host} "test -f {done_file} && echo done"'
    try:
        result = subprocess.run(
            cmd, shell=True, capture_output=True, text=True, timeout=30
        )
        return result.stdout.strip() == 'done'
    except subprocess.TimeoutExpired:
        print(f"[auto_pull] SSH timeout checking {host}:{done_file}")
        return False
    except Exception as e:
        print(f"[auto_pull] Error checking {host}:{done_file}: {e}")
        return False


def pull_results(host: str, remote_log_dir: str, local_log_dir: str, bare: bool = False) -> bool:
    """Pull results from remote host to local directory."""
    # Create local directory if it doesn't exist
    os.makedirs(os.path.dirname(local_log_dir), exist_ok=True)

    cmd = f'rsync -avzh --progress {host}:{remote_log_dir}/ {local_log_dir}/'

    if bare:
        # Exclude large files
        cmd += " --exclude '*.pkl' --exclude '*.png' --exclude '*.gif' --exclude '*.pth' --exclude '*.pt'"

    print(f"[auto_pull] Pulling: {host}:{remote_log_dir} -> {local_log_dir}")
    try:
        result = subprocess.run(cmd, shell=True, timeout=600)  # 10 min timeout
        return result.returncode == 0
    except subprocess.TimeoutExpired:
        print(f"[auto_pull] Rsync timeout for {host}:{remote_log_dir}")
        return False
    except Exception as e:
        print(f"[auto_pull] Error pulling from {host}:{remote_log_dir}: {e}")
        return False


def load_manifest(manifest_path: str) -> list:
    """Load job manifest from JSON file."""
    if not os.path.exists(manifest_path):
        return []
    with open(manifest_path, 'r') as f:
        return json.load(f)


def save_manifest(manifest_path: str, jobs: list):
    """Save job manifest to JSON file."""
    os.makedirs(os.path.dirname(manifest_path), exist_ok=True)
    with open(manifest_path, 'w') as f:
        json.dump(jobs, f, indent=2)


def poll_and_pull(manifest_path: str, poll_interval: int = 60, bare: bool = False, once: bool = False):
    """
    Main polling loop. Checks for .done markers and pulls completed jobs.

    Args:
        manifest_path: Path to the job manifest JSON file
        poll_interval: Seconds between polls
        bare: If True, exclude large files when pulling
        once: If True, run once and exit instead of looping
    """
    print(f"[auto_pull] Starting poller, manifest: {manifest_path}")
    print(f"[auto_pull] Poll interval: {poll_interval}s, bare mode: {bare}")

    while True:
        jobs = load_manifest(manifest_path)

        if not jobs:
            if once:
                print("[auto_pull] No jobs in manifest, exiting")
                break
            time.sleep(poll_interval)
            continue

        pending_jobs = [j for j in jobs if j.get('status') == 'pending']

        if not pending_jobs:
            if once:
                print("[auto_pull] No pending jobs, exiting")
                break
            print(f"[auto_pull] No pending jobs, sleeping {poll_interval}s")
            time.sleep(poll_interval)
            continue

        print(f"[auto_pull] Checking {len(pending_jobs)} pending jobs...")

        for job in pending_jobs:
            host = job['host']
            remote_log_dir = job['remote_log_dir']
            local_log_dir = job['local_log_dir']
            exp_name = job.get('exp_name', 'unknown')

            if check_done_marker(host, remote_log_dir):
                print(f"[auto_pull] Job completed: {exp_name} on {host}")

                if pull_results(host, remote_log_dir, local_log_dir, bare=bare):
                    job['status'] = 'pulled'
                    job['pulled_at'] = datetime.now().isoformat()
                    print(f"[auto_pull] Successfully pulled: {local_log_dir}")
                else:
                    job['status'] = 'pull_failed'
                    print(f"[auto_pull] Failed to pull: {exp_name}")

                # Save manifest after each job status change
                save_manifest(manifest_path, jobs)

        if once:
            break

        time.sleep(poll_interval)


def main():
    parser = argparse.ArgumentParser(description='Chester Auto Pull - Background result synchronization')
    parser.add_argument('--manifest', type=str, required=True,
                        help='Path to job manifest JSON file')
    parser.add_argument('--poll-interval', type=int, default=60,
                        help='Seconds between polls (default: 60)')
    parser.add_argument('--bare', action='store_true',
                        help='Exclude large files (*.pkl, *.pth, etc.)')
    parser.add_argument('--once', action='store_true',
                        help='Run once and exit instead of continuous polling')

    args = parser.parse_args()

    try:
        poll_and_pull(
            manifest_path=args.manifest,
            poll_interval=args.poll_interval,
            bare=args.bare,
            once=args.once
        )
    except KeyboardInterrupt:
        print("\n[auto_pull] Interrupted by user")
        sys.exit(0)


if __name__ == '__main__':
    main()



"""
Utilities for integrating Chester with Hydra.

Note: hydra-core is an optional dependency. Functions that require hydra
will raise ImportError if hydra is not installed.
"""

import os
import os.path as osp
from typing import Dict, Any, List, Union, Callable
from pathlib import Path
import base64
import pickle

# Lazy imports for optional hydra dependency
_hydra = None
_HydraConfig = None
_read_write = None
_open_dict = None


def _require_hydra():
    """Import hydra and related modules. Raises ImportError if not available."""
    global _hydra, _HydraConfig, _read_write, _open_dict
    if _hydra is None:
        try:
            import hydra
            from hydra.core.hydra_config import HydraConfig
            from omegaconf import read_write, open_dict
            _hydra = hydra
            _HydraConfig = HydraConfig
            _read_write = read_write
            _open_dict = open_dict
        except ImportError:
            raise ImportError(
                "hydra-core is required for this function. "
                "Install it with: pip install hydra-core"
            )
    return _hydra, _HydraConfig, _read_write, _open_dict

def _format_hydra_value(value: Any) -> str:
    """Format a value for Hydra command line override."""
    if isinstance(value, bool):
        return str(value).lower()
    elif isinstance(value, (int, float)):
        return str(value)
    elif isinstance(value, str):
        # If it contains spaces, quote it
        if ' ' in value:
            return f'"{value}"'
        return value
    elif isinstance(value, list):
        # Format list values with Hydra's format [val1,val2,...]
        formatted_items = []
        for item in value:
            if isinstance(item, str) and ',' in item:
                # Escape commas in string values
                formatted_items.append(f'"{item}"')
            else:
                formatted_items.append(_format_hydra_value(item))
        return f"[{','.join(formatted_items)}]"
    return str(value)


def variant_to_hydra_overrides(variant: Dict[str, Any]) -> List[str]:
    """
    Convert a chester variant dictionary to Hydra command line overrides.
    
    Args:
        variant: Dictionary of parameters from chester's VariantGenerator
        
    Returns:
        List of strings, each being a Hydra command line override (e.g., ['key1=val1', 'key2=val2', ...])
    """
    overrides = []
    
    for key, value in variant.items():
        # Skip special chester keys
        if key in ['chester_first_variant', 'chester_last_variant', 'is_debug']:
            continue
        
        # Handle nested dictionaries
        if isinstance(value, dict):
            for nested_key, nested_value in value.items():
                if str(nested_key) == "_name":
                    # Special case for _name which becomes just the parent key in Hydra
                    overrides.append(f"{key}={_format_hydra_value(nested_value)}")
                else:
                    overrides.append(f"{key}.{nested_key}={_format_hydra_value(nested_value)}")
        else:
            overrides.append(f"{key}={_format_hydra_value(value)}")
    
    return overrides


def to_hydra_command(
    params: Dict[str, Any],
    python_command: str = "python",
    script: str = "main_df.py",
    hydra_flags: Dict[str, Any] = None,
    env: Dict[str, Any] = {},
) -> str:
    """
    Convert parameters to a Hydra command.
    
    Args:
        params: Dictionary of parameters
        python_command: Python command to use
        script: Script to run
        hydra_flags: Additional Hydra flags (e.g., {'multirun': True})
        
    Returns:
        Command string to execute
    """
    variant_data = pickle.loads(base64.b64decode(params.pop("variant_data")))
    variant_data["hydra.run.dir"] = params["log_dir"]
    # Extract variant_data to convert to overrides
    command = python_command + " " + script
    for k, v in env.items():
        command = ("%s=%s " % (k, v)) + command
    # Add Hydra flags
    if hydra_flags:
        for flag, value in hydra_flags.items():
            if value is True:
                command += f" --{flag}"
            elif value not in [False, None]:
                command += f" --{flag}={value}"
    
    # Add Hydra overrides
    overrides = variant_to_hydra_overrides(variant_data)
    if overrides:
        command += " " + " ".join(overrides)
    
    return command 

def run_hydra_command(command: str, log_dir: str, stub_method_call: Callable):
    hydra, HydraConfig, read_write, open_dict = _require_hydra()

    cmd_parts = command.split()
    # Find the index where the python module starts (after python command and potential -m flag)
    module_start_idx = 0
    for i, part in enumerate(cmd_parts):
        if part.endswith('.py') or (part == '-m' and i+1 < len(cmd_parts)):
            module_start_idx = i + 2 if part == '-m' else i + 1
            break

    # Everything after the python module are hydra overrides
    overrides = cmd_parts[module_start_idx:]
    with hydra.initialize(config_path="../configs"):
        cfg = hydra.compose(config_name="config",
                            overrides=overrides,
                            return_hydra_config=True)
        # output_dir = cfg.hydra.run.dir
        output_dir = log_dir
        with read_write(cfg.hydra.runtime):
            with open_dict(cfg.hydra.runtime):
                cfg.hydra.runtime.output_dir = os.path.abspath(output_dir)
        Path(str(output_dir)).mkdir(parents=True, exist_ok=True)
        HydraConfig.instance().set_config(cfg)
        stub_method_call(cfg)


import os
import sys
import shutil
import os.path as osp
import json
import time
import datetime
import dateutil.tz
import tempfile
from collections import defaultdict

# LOG_OUTPUT_FORMATS = ['stdout', 'log', 'csv', 'tensorboard']
LOG_OUTPUT_FORMATS = ['stdout', 'log', 'csv']
# Also valid: json, tensorboard

DEBUG = 10
INFO = 20
WARN = 30
ERROR = 40

DISABLED = 50


class KVWriter(object):
    def writekvs(self, kvs):
        raise NotImplementedError


class SeqWriter(object):
    def writeseq(self, seq):
        raise NotImplementedError


def put_in_middle(str1, str2):
    # Put str1 in str2
    n = len(str1)
    m = len(str2)
    if n <= m:
        return str2
    else:
        start = (n - m) // 2
        return str1[:start] + str2 + str1[start + m:]


class HumanOutputFormat(KVWriter, SeqWriter):
    def __init__(self, filename_or_file):
        if isinstance(filename_or_file, str):
            self.file = open(filename_or_file, 'wt')
            self.own_file = True
        else:
            assert hasattr(filename_or_file, 'read'), 'expected file or str, got %s' % filename_or_file
            self.file = filename_or_file
            self.own_file = False

    def writekvs(self, kvs):
        # Create strings for printing
        key2str = {}
        for (key, val) in sorted(kvs.items()):
            if isinstance(val, float):
                valstr = '%-8.3g' % (val,)
            else:
                valstr = str(val)
            key2str[self._truncate(key)] = self._truncate(valstr)

        # Find max widths
        if len(key2str) == 0:
            print('WARNING: tried to write empty key-value dict')
            return
        else:
            keywidth = max(map(len, key2str.keys()))
            valwidth = max(map(len, key2str.values()))

        # Write out the data
        now = datetime.datetime.now(dateutil.tz.tzlocal())
        timestamp = now.strftime('%Y-%m-%d %H:%M:%S.%f %Z')

        dashes = '-' * (keywidth + valwidth + 7)
        dashes_time = put_in_middle(dashes, timestamp)
        lines = [dashes_time]
        for (key, val) in sorted(key2str.items()):
            lines.append('| %s%s | %s%s |' % (
                key,
                ' ' * (keywidth - len(key)),
                val,
                ' ' * (valwidth - len(val)),
            ))
        lines.append(dashes)
        self.file.write('\n'.join(lines) + '\n')

        # Flush the output to the file
        self.file.flush()

    def _truncate(self, s):
        return s[:30] + '...' if len(s) > 33 else s

    def writeseq(self, seq):
        for arg in seq:
            self.file.write(arg)
        self.file.write('\n')
        self.file.flush()

    def close(self):
        if self.own_file:
            self.file.close()


class JSONOutputFormat(KVWriter):
    def __init__(self, filename):
        self.file = open(filename, 'wt')

    def writekvs(self, kvs):
        for k, v in sorted(kvs.items()):
            if hasattr(v, 'dtype'):
                v = v.tolist()
                kvs[k] = float(v)
        self.file.write(json.dumps(kvs) + '\n')
        self.file.flush()

    def close(self):
        self.file.close()


class CSVOutputFormat(KVWriter):
    def __init__(self, filename):
        self.file = open(filename, 'w+t')
        self.keys = []
        self.sep = ','

    def writekvs(self, kvs):
        # Add our current row to the history
        extra_keys = kvs.keys() - self.keys
        if extra_keys:
            self.keys.extend(extra_keys)
            self.file.seek(0)
            lines = self.file.readlines()
            self.file.seek(0)
            for (i, k) in enumerate(self.keys):
                if i > 0:
                    self.file.write(',')
                self.file.write(k)
            self.file.write('\n')
            for line in lines[1:]:
                self.file.write(line[:-1])
                self.file.write(self.sep * len(extra_keys))
                self.file.write('\n')
        for (i, k) in enumerate(self.keys):
            if i > 0:
                self.file.write(',')
            v = kvs.get(k)
            if v is not None:
                self.file.write(str(v))
        self.file.write('\n')
        self.file.flush()

    def close(self):
        self.file.close()


class TensorBoardOutputFormat(KVWriter):
    """
    Dumps key/value pairs into TensorBoard's numeric format.
    """

    def __init__(self, dir):
        os.makedirs(dir, exist_ok=True)
        self.dir = dir
        self.step = 1
        prefix = 'events'
        path = osp.join(osp.abspath(dir), prefix)
        import tensorflow as tf
        from tensorflow.python import pywrap_tensorflow
        from tensorflow.core.util import event_pb2
        from tensorflow.python.util import compat
        self.tf = tf
        self.event_pb2 = event_pb2
        self.pywrap_tensorflow = pywrap_tensorflow
        self.writer = pywrap_tensorflow.EventsWriter(compat.as_bytes(path))

    def writekvs(self, kvs):
        def summary_val(k, v):
            kwargs = {'tag': k, 'simple_value': float(v)}
            return self.tf.Summary.Value(**kwargs)

        summary = self.tf.Summary(value=[summary_val(k, v) for k, v in kvs.items()])
        event = self.event_pb2.Event(wall_time=time.time(), summary=summary)
        event.step = self.step  # is there any reason why you'd want to specify the step?
        self.writer.WriteEvent(event)
        self.writer.Flush()
        self.step += 1

    def close(self):
        if self.writer:
            self.writer.Close()
            self.writer = None


def make_output_format(format, ev_dir, log_suffix=''):
    os.makedirs(ev_dir, exist_ok=True)
    if format == 'stdout':
        return HumanOutputFormat(sys.stdout)
    elif format == 'log':
        return HumanOutputFormat(osp.join(ev_dir, 'log%s.txt' % log_suffix))
    elif format == 'json':
        return JSONOutputFormat(osp.join(ev_dir, 'progress%s.json' % log_suffix))
    elif format == 'csv':
        return CSVOutputFormat(osp.join(ev_dir, 'progress%s.csv' % log_suffix))
    elif format == 'tensorboard':
        return TensorBoardOutputFormat(osp.join(ev_dir, 'tb%s' % log_suffix))
    else:
        raise ValueError('Unknown format specified: %s' % (format,))


# ================================================================
# API
# ================================================================

def logkv(key, val):
    """
    Log a value of some diagnostic
    Call this once for each diagnostic quantity, each iteration
    If called many times, last value will be used.
    """
    Logger.CURRENT.logkv(key, val)


def logkv_mean(key, val):
    """
    The same as logkv(), but if called many times, values averaged.
    """
    Logger.CURRENT.logkv_mean(key, val)


def logkvs(d):
    """
    Log a dictionary of key-value pairs
    """
    for (k, v) in d.items():
        logkv(k, v)


def dumpkvs():
    """
    Write all of the diagnostics from the current iteration

    level: int. (see logger.py docs) If the global logger level is higher than
                the level argument here, don't print to stdout.
    """
    Logger.CURRENT.dumpkvs()


def getkvs():
    return Logger.CURRENT.name2val


def log(*args, level=INFO):
    """
    Write the sequence of args, with no separators, to the console and output files (if you've configured an output file).
    """
    Logger.CURRENT.log(*args, level=level)


def debug(*args):
    log(*args, level=DEBUG)


def info(*args):
    log(*args, level=INFO)


def warn(*args):
    log(*args, level=WARN)


def error(*args):
    log(*args, level=ERROR)


def set_level(level):
    """
    Set logging threshold on current logger.
    """
    Logger.CURRENT.set_level(level)


def get_dir():
    """
    Get directory that log files are being written to.
    will be None if there is no output directory (i.e., if you didn't call start)
    """
    return Logger.CURRENT.get_dir()


record_tabular = logkv
dump_tabular = dumpkvs


class ProfileKV:
    """
    Usage:
    with logger.ProfileKV("interesting_scope"):
        code
    """

    def __init__(self, n):
        self.n = "wait_" + n

    def __enter__(self):
        self.t1 = time.time()

    def __exit__(self, type, value, traceback):
        Logger.CURRENT.name2val[self.n] += time.time() - self.t1


def profile(n):
    """
    Usage:
    @profile("my_func")
    def my_func(): code
    """

    def decorator_with_name(func):
        def func_wrapper(*args, **kwargs):
            with ProfileKV(n):
                return func(*args, **kwargs)

        return func_wrapper

    return decorator_with_name


# ================================================================
# Backend
# ================================================================

class Logger(object):
    DEFAULT = None  # A logger with no output files. (See right below class definition)
    # So that you can still log to the terminal without setting up any output files
    CURRENT = None  # Current logger being used by the free functions above

    def __init__(self, dir, output_formats):
        self.name2val = defaultdict(float)  # values this iteration
        self.name2cnt = defaultdict(int)
        self.level = INFO
        self.dir = dir
        self.output_formats = output_formats

    # Logging API, forwarded
    # ----------------------------------------
    def logkv(self, key, val):
        self.name2val[key] = val

    def logkv_mean(self, key, val):
        if val is None:
            self.name2val[key] = None
            return
        oldval, cnt = self.name2val[key], self.name2cnt[key]
        self.name2val[key] = oldval * cnt / (cnt + 1) + val / (cnt + 1)
        self.name2cnt[key] = cnt + 1

    def dumpkvs(self):
        if self.level == DISABLED: return
        for fmt in self.output_formats:
            if isinstance(fmt, KVWriter):
                fmt.writekvs(self.name2val)
        self.name2val.clear()
        self.name2cnt.clear()

    def log(self, *args, level=INFO):
        if self.level <= level:
            self._do_log(args)

    # Configuration
    # ----------------------------------------
    def set_level(self, level):
        self.level = level

    def get_dir(self):
        return self.dir

    def close(self):
        for fmt in self.output_formats:
            fmt.close()

    # Misc
    # ----------------------------------------
    def _do_log(self, args):
        for fmt in self.output_formats:
            if isinstance(fmt, SeqWriter):
                fmt.writeseq(map(str, args))


Logger.DEFAULT = Logger.CURRENT = Logger(dir=None, output_formats=[HumanOutputFormat(sys.stdout)])


def configure(dir=None, format_strs=None, exp_name=None):
    if dir is None:
        dir = os.getenv('OPENAI_LOGDIR')
    if dir is None:
        dir = osp.join(tempfile.gettempdir(),
                       datetime.datetime.now().strftime("chester-%Y-%m-%d-%H-%M-%S"))

    assert isinstance(dir, str)
    os.makedirs(dir, exist_ok=True)

    if format_strs is None:
        strs = os.getenv('OPENAI_LOG_FORMAT')
        format_strs = strs.split(',') if strs else LOG_OUTPUT_FORMATS
    output_formats = [make_output_format(f, dir) for f in format_strs]

    Logger.CURRENT = Logger(dir=dir, output_formats=output_formats)
    log('Logging to %s' % dir)


def reset():
    if Logger.CURRENT is not Logger.DEFAULT:
        Logger.CURRENT.close()
        Logger.CURRENT = Logger.DEFAULT
        log('Reset logger')


class scoped_configure(object):
    def __init__(self, dir=None, format_strs=None):
        self.dir = dir
        self.format_strs = format_strs
        self.prevlogger = None

    def __enter__(self):
        self.prevlogger = Logger.CURRENT
        configure(dir=self.dir, format_strs=self.format_strs)

    def __exit__(self, *args):
        Logger.CURRENT.close()
        Logger.CURRENT = self.prevlogger


# ================================================================

def _demo():
    info("hi")
    debug("shouldn't appear")
    set_level(DEBUG)
    debug("should appear")
    dir = "/tmp/testlogging"
    if os.path.exists(dir):
        shutil.rmtree(dir)
    configure(dir=dir)
    logkv("a", 3)
    logkv("b", 2.5)
    dumpkvs()
    logkv("b", -2.5)
    logkv("a", 5.5)
    dumpkvs()
    info("^^^ should see a = 5.5")
    logkv_mean("b", -22.5)
    logkv_mean("b", -44.4)
    logkv("a", 5.5)
    dumpkvs()
    info("^^^ should see b = 33.3")

    logkv("b", -2.5)
    dumpkvs()

    logkv("a", "longasslongasslongasslongasslongasslongassvalue")
    dumpkvs()


# ================================================================
# Readers
# ================================================================

def read_json(fname):
    import pandas
    ds = []
    with open(fname, 'rt') as fh:
        for line in fh:
            ds.append(json.loads(line))
    return pandas.DataFrame(ds)


def read_csv(fname):
    import pandas
    return pandas.read_csv(fname, index_col=None, comment='#')


def read_tb(path):
    """
    path : a tensorboard file OR a directory, where we will find all TB files
           of the form events.*
    """
    import pandas
    import numpy as np
    from glob import glob
    from collections import defaultdict
    import tensorflow as tf
    if osp.isdir(path):
        fnames = glob(osp.join(path, "events.*"))
    elif osp.basename(path).startswith("events."):
        fnames = [path]
    else:
        raise NotImplementedError("Expected tensorboard file or directory containing them. Got %s" % path)
    tag2pairs = defaultdict(list)
    maxstep = 0
    for fname in fnames:
        for summary in tf.train.summary_iterator(fname):
            if summary.step > 0:
                for v in summary.summary.value:
                    pair = (summary.step, v.simple_value)
                    tag2pairs[v.tag].append(pair)
                maxstep = max(summary.step, maxstep)
    data = np.empty((maxstep, len(tag2pairs)))
    data[:] = np.nan
    tags = sorted(tag2pairs.keys())
    for (colidx, tag) in enumerate(tags):
        pairs = tag2pairs[tag]
        for (step, value) in pairs:
            data[step - 1, colidx] = value
    return pandas.DataFrame(data, columns=tags)


if __name__ == "__main__":
    _demo()



# The scheduler sweeps over available GPUs for the nodes  in the CHESTER_QUEUE_DIR
import os
import glob
import pdb
import pickle
import time
import sys

from chester import config
from chester.utils_logger import timelog
import psutil

check_interval = 120  # Check every 60 seconds for available GPUs
user_name = 'zixuanhu'


def checkIfProcessRunning(processName):
    '''
    Check if there is any running process that contains the given name processName.
    '''
    # Iterate over the all the running process
    for proc in psutil.process_iter():
        try:
            # Check if process name contains the given name string.
            if processName.lower() in proc.name().lower() and user_name == proc.username():
                return True
        except (psutil.NoSuchProcess, psutil.AccessDenied, psutil.ZombieProcess):
            pass
    return False


def check_available_nodes():
    all_stats_path = glob.glob(config.GPU_STATE_DIR + '/*.pkl')
    stats_names = [p.split('/')[-1][:-4] for p in all_stats_path]
    all_stats = []
    for p in all_stats_path:
        while 1:
            try:

                all_stats.append(pickle.load(open(p, 'rb')))
                time.sleep(0.1)
            except:
                print(p, ' is broken')
                continue
            break
    # all_stats = [ for p in all_stats_path]
    available_nodes = {}
    for name, (sys_data, gpu_data) in zip(stats_names, all_stats):
        gpu_ids = []
        for i, data in enumerate(gpu_data):
            if len(data['procs']) == 0:  # GPU is available if no user process is found on the GPU
                # if data['mem_usage'] < 5 \
                #     or (len(data['procs']) > 0 and data['procs'][0][1] == 'zixuanhu' and data['mem_usage'] < 40):
                gpu_ids.append((i, data['mem_free']))
        if len(gpu_ids) > 0:
            available_nodes[name] = gpu_ids
    return available_nodes


if checkIfProcessRunning('remote_scheduler'):
    exit()

if __name__ == '__main__':
    print('cur folder ', os.getcwd())
    while 1:
        # check jobs in the queue
        tasks = glob.glob(os.path.join(config.CHESTER_QUEUE_DIR, 'queues', '*'))
        print(f' {len(tasks)} jobs to be scheduled')
        tasks_with_time = [(os.path.getmtime(t), t) for t in tasks if os.path.isfile(t)]
        sorted_tasks = sorted(tasks_with_time)
        # check if any GPUs are available
        available_GPUs = check_available_nodes()  # Dictionary: {node_name:[available_gpu_id],...]
        # available_GPUs.pop('autobot-0-17', None)
        # available_GPUs.pop('autobot-0-25', None)
        # available_GPUs.pop('autobot-0-33', None)
        # available_GPUs.pop('autobot-0-37', None)
        # available_GPUs.pop('autobot-0-23', None)
        # available_GPUs["autobot-1-1"] = [(0, 10000), (1, 10000),
        #                                  (2, 10000), (3, 10000)]
        short_list = [
            # "autobot-0-25",
            "autobot-0-29",
            "autobot-0-33",
            "autobot-0-37",
        ]
        # for node in short_list:
        #     if node in available_GPUs:
        #         available_GPUs[node] = available_GPUs[node][:6]
        short_list = [
            # "autobot-0-25",
            'autobot-0-9',
            'autobot-0-11',
            "autobot-0-13", "autobot-0-15", "autobot-0-17", "autobot-0-19", "autobot-0-21",
            'autobot-0-23',  # 4x2080 + 32 cpus
        ]
        for node in short_list:
            if node in available_GPUs:
                available_GPUs[node] = available_GPUs[node][:3]
        short_list = [
            "autobot-1-1", "autobot-1-6",
        ]
        for node in short_list:
            if node in available_GPUs:
                available_GPUs[node] = available_GPUs[node][:4]
        # available_GPUs = {
        #     # 'autobot-0-9': [0, 1],
        #     # 'autobot-0-11': [0, 1],
        #     # 'autobot-0-13': [0, 1],
        #     # 'autobot-0-15': [0, 1, 2, 3],
        #     'autobot-0-15': [0, 1],
        #     # 'autobot-0-17': [0, 1, 2, 3],
        #     'autobot-0-19': [0, 1],
        #     'autobot-0-21': [0, 1],
        #     # 'autobot-0-23': [0, 1],
        #     # 'autobot-0-25': [ 1, 2, 3, 4, 5, 6, 7],
        #     # 'autobot-0-29': [0, 1, 2, 3, 4, 5, 6, 7],
        #
        #     'autobot-1-1': [0, 1],
        #     'autobot-1-6': [0, 1],
        #     'autobot-1-10': [0, 1, 2, 3, 4, 5, 6, 7, 8],
        # }  # Temporary
        # timelog('Available GPUs: ' + str(available_GPUs))
        for node, gpus in available_GPUs.items():
            print(f"##### {node}")
            for i, mem in gpus:
                print(f'    GPU {i}: {mem} MB')

        succ_tasks = 0
        for _, script in sorted_tasks:
            with open(script) as f:
                while True:  # Read header files
                    line = f.readline()
                    if line[0] != '#':
                        break
                    header = line.split(' ')[0][1:]
                    if header == 'CHESTERNODE':
                        node_list = line.rstrip()[13:].split(',')
                    elif header == 'CHESTEROUT':
                        stdout_file = line.rstrip().split(' ')[1]
                    elif header == 'CHESTERERR':
                        stderr_file = line.rstrip().split(' ')[1]
                    elif header == 'CHESTERSCRIPT':
                        script_file = line.rstrip().split(' ')[1]
            for node in node_list:
                real_node = 'autobot-' + node
                # TODO: add gpu num
                if real_node in available_GPUs:
                    gpu_ids = available_GPUs[real_node]
                    if len(gpu_ids) > 0:
                        env_command = f'CUDA_VISIBLE_DEVICES={gpu_ids[0][0]} '  # CUDA_LAUNCH_BLOCKING=1 '
                        command = f"ssh -q {real_node} \'{env_command} bash {script_file} </dev/null >{stdout_file} 2>{stderr_file} &\'"
                        # command = f"ssh -q {real_node} \'{env_command} python /home/xlin3/test.py </dev/null >{stdout_file} 2>{stderr_file} &\'"
                        rm_command = f'rm {script}'
                        timelog(f"Job launched on node {real_node}, GPU {gpu_ids[0]}, {script_file}")
                        os.system(command)
                        os.system(rm_command)
                        gpu_ids.pop(0)
                        succ_tasks += 1
                        break
                    else:
                        del available_GPUs[real_node]
        if succ_tasks == len(sorted_tasks):
            timelog(f'All {succ_tasks} jobs done!')
            print("==================== ", flush=True)
            last_log = sorted(glob.glob('/home/zixuanhu/chester_scheduler/logs/*'))[-1]
            dir_name = os.path.dirname(last_log)
            os.system(f'cp {last_log}  {dir_name}/last_log.txt')
            break
        time.sleep(check_interval)



import os.path as osp
import os

PROJECT_PATH = osp.abspath(osp.join(osp.dirname(__file__), '..'))

AWS_NETWORK_INTERFACES = []

AWS_BUCKET_REGION_NAME = 'us-east-2'

MUJOCO_KEY_PATH = osp.expanduser("~/.mujoco")

USE_GPU = True

USE_TF = True

AWS_REGION_NAME = "us-east-2"

if USE_GPU:
    DOCKER_IMAGE = "dementrock/rllab3-shared-gpu"
else:
    DOCKER_IMAGE = "dementrock/rllab3-shared"

DOCKER_LOG_DIR = "/tmp/expt"

CODE_DIR = "/root/code/"

AWS_S3_PATH = "s3://chester-softgym/rllab/experiments"

EBS_OPTIMIZED = True

AWS_EXTRA_CONFIGS = dict()

AWS_CODE_SYNC_S3_PATH = "s3://chester-softgym/rllab/code"

ALL_REGION_AWS_IMAGE_IDS = {
    # "ap-northeast-1": "ami-002f0167",
    # "ap-northeast-2": "ami-590bd937",
    # "ap-south-1": "ami-77314318",
    # "ap-southeast-1": "ami-1610a975",
    # "ap-southeast-2": "ami-9dd4ddfe",
    # "eu-central-1": "ami-63af720c",
    # "eu-west-1": "ami-41484f27",
    # "sa-east-1": "ami-b7234edb",
    "us-east-1": "ami-83f26195",
    "us-east-2": "ami-0ec385d5f98faacc3",   #"ami-0e63a1a8842443350",
    "us-west-1": "ami-576f4b37",
    "us-west-2": "ami-b8b62bd8"
}

AWS_IMAGE_ID = ALL_REGION_AWS_IMAGE_IDS[AWS_REGION_NAME]

if USE_GPU:
    AWS_INSTANCE_TYPE = "p2.xlarge"
else:
    AWS_INSTANCE_TYPE = "c4.4xlarge"

ALL_REGION_AWS_KEY_NAMES = {
    "us-east-1": "rllab-us-east-1",
    "us-east-2": "rllab-us-east-2",
    "us-west-1": "rllab-us-west-1",
    "us-west-2": "rllab-us-west-2"
}

AWS_KEY_NAME = ALL_REGION_AWS_KEY_NAMES[AWS_REGION_NAME]

AWS_SPOT = True

AWS_SPOT_PRICE = '2.0'

AWS_ACCESS_KEY = os.environ.get("AWS_ACCESS_KEY", None)

AWS_ACCESS_SECRET = os.environ.get("AWS_ACCESS_SECRET", None)

AWS_IAM_INSTANCE_PROFILE_NAME = "rllab"

AWS_SECURITY_GROUPS = ["rllab-sg"]

ALL_REGION_AWS_SECURITY_GROUP_IDS = {
    "us-east-1": [
        "sg-e17bfc95"
    ],
    "us-east-2": [
        "sg-1ddb3876"
    ],
    "us-west-1": [
        "sg-cd5f9db4"
    ],
    "us-west-2": [
        "sg-b585a8c9"
    ]
}

AWS_SECURITY_GROUP_IDS = ALL_REGION_AWS_SECURITY_GROUP_IDS[AWS_REGION_NAME]

FAST_CODE_SYNC_IGNORES = [
    ".git",
    "data/autobot",
    "data/corl_s3_data",
    "data/videos",
    "data/open_loop_videos",
    "data/icml",
    "data/local",
    "data/seuss",
    "data/yufei_s3_data",
    "data/icml"
    "data/local",
    "data/archive",
    "data/debug",
    "data/s3",
    "data/video",
    ".idea",
    "tests",
    "examples",
    "docs",
    ".idea",
    ".DS_Store",
    ".ipynb_checkpoints",
    "blackbox",
    "blackbox.zip",
    "*.pyc",
    "*.ipynb",
    "scratch-notebooks",
    "conopt_root",
    "private/key_pairs",
    "DPI-Net",
    "imgs/",
    "imgs",
    "videos"
]

FAST_CODE_SYNC = True

LABEL = ""



import os

from chester.autobot_config import AUTOBOT_NODELIST

for node in AUTOBOT_NODELIST:
    real_node = 'autobot-' + node
    print("killing ", real_node)
    command = f'ssh {real_node} "pkill -9 -u zixuanhu"'

    os.system(command)
